import logging
import os
import sys

from werkzeug.debug import DebuggedApplication

logging.basicConfig(
  level=logging.DEBUG,
  format="%(levelname)s: %(asctime)s {%(filename)s:%(lineno)d}: %(message)s "
)

(file_path, fname) = os.path.split(__file__)
app_path = os.path.dirname(file_path)
os.chdir(app_path)
logging.info("My path is " + app_path)
sys.path.insert (0, app_path)
sys.stdout = sys.stderr
from sanskrit_parser.rest_api import run

application = DebuggedApplication(run.app, True)



#!/usr/bin/env python
from __future__ import print_function
from indic_transliteration import sanscript
from indic_transliteration import detect
from sanskrit_parser.util import normalization

from contextlib import contextmanager
import logging
import six

# Wrap scheme names defined in sanscript
BENGALI = sanscript.BENGALI
DEVANAGARI = sanscript.DEVANAGARI
GUJARATI = sanscript.GUJARATI
GURMUKHI = sanscript.GURMUKHI
KANNADA = sanscript.KANNADA
MALAYALAM = sanscript.MALAYALAM
ORIYA = sanscript.ORIYA
TAMIL = sanscript.TAMIL
TELUGU = sanscript.TELUGU
HK = sanscript.HK
IAST = sanscript.IAST
ITRANS = sanscript.ITRANS
KOLKATA = sanscript.KOLKATA
SLP1 = sanscript.SLP1
VELTHUIS = sanscript.VELTHUIS
WX = sanscript.WX

# Dict defined so autodetect can work
SCHEMES = {
    'Bengali': BENGALI,
    'Devanagari': DEVANAGARI,
    'Gujarati': GUJARATI,
    'Gurmukhi': GURMUKHI,
    'Kannada': KANNADA,
    'Malayalam': MALAYALAM,
    'Oriya': ORIYA,
    'Tamil': TAMIL,
    'Telugu': TELUGU,
    'HK': HK,
    'IAST': IAST,
    'ITRANS': ITRANS,
    'Kolkata': KOLKATA,
    'SLP1': SLP1,
    'Velthuis': VELTHUIS,
    'WX': WX
}

logger = logging.getLogger(__name__)
denormalize = False


class SanskritString(object):
    """ Sanskrit String Class: Base of the class hierarchy

        Attributes:
           thing(str)   : thing to be represented
           encoding(str): SanskritBase encoding of thing as passed (eg: SanskritBase.HK, SanskritBase.DEVANAGARI)
        Args:
           thing(str):    As above
           encoding(str): As above

    """

    def __init__(self, thing, encoding=None, unicode_encoding='utf-8'):
        assert isinstance(thing, six.string_types)
        # Encode early, unicode everywhere, decode late is the philosophy
        # However, we need to accept both unicode and non unicode strings
        # We are udAramatiH
        if isinstance(thing, six.text_type):
            self.thing = thing
        else:
            self.thing = six.text_type(thing, unicode_encoding)
        if encoding is None:
            # Autodetect Encoding
            encoding = SCHEMES[detect.detect(self.thing)]
        if encoding != SLP1:
            # Convert to SLP1
            self.thing = sanscript.transliterate(self.thing, encoding, SLP1)
            # At this point, we are guaranteed that internal
            # representation is in SLP1

    def transcoded(self, encoding=None, strict_io=True):
        """ Return a transcoded version of self

            Args:
              encoding(SanskritObject.Scheme):
            Returns:
              str: transcoded version
        """
        s = self.thing
        if not strict_io:
            s = normalization.denormalize(s)
        return sanscript.transliterate(s, SLP1, encoding)

    def canonical(self, strict_io=True):
        """ Return canonical transcoding (SLP1) of self
        """
        return self.transcoded(SLP1, strict_io)

    def devanagari(self, strict_io=True):
        """ Return devanagari transcoding of self
        """
        return self.transcoded(DEVANAGARI, strict_io)

    def __str__(self):
        global denormalize
        s = self.transcoded(SLP1)
        if denormalize:
            s = normalization.denormalize(s)
        return s

    def __repr__(self):
        return str(self)


class SanskritImmutableString(SanskritString):
    """ Immutable version of SanskritString
    """
    def __init__(self, thing=None, encoding=None, unicode_encoding='utf-8'):
        super().__init__(thing, encoding, unicode_encoding)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)


class SanskritNormalizedString(SanskritString):
    """ SanskritString plus Normalization of input
    """
    def __init__(self, thing=None, encoding=None, unicode_encoding='utf-8',
                 strict_io=True, replace_ending_visarga='s'):
        super().__init__(thing, encoding, unicode_encoding)
        if not strict_io:
            # Normalize
            logger.debug("Before normalization: %s", self.thing)
            tmp = normalization.normalize(self.thing)
            if replace_ending_visarga == 's':
                self.thing = normalization.replace_ending_visarga_s(tmp)
            elif replace_ending_visarga == 'r':
                self.thing = normalization.replace_ending_visarga_r(tmp)
            else:
                self.thing = tmp
            # Lazy Anusvaras (see issue #103)
            try:
                self.thing = sanscript.SCHEMES[sanscript.SLP1].fix_lazy_anusvaara(self.thing)
            except (NameError, AttributeError):
                print("Not fixing lazy anusvaras, you probably have an older version of indic_transliteration")
            logger.debug("After normalization: %s", self.thing)


class SanskritObject(SanskritNormalizedString):
    """ Sanskrit Object Class: Derived From SanskritString

        Attributes:

    """
    def __init__(self, thing=None, encoding=None, unicode_encoding='utf-8',
                 strict_io=True, replace_ending_visarga='s'):
        super().__init__(thing, encoding, unicode_encoding, strict_io, replace_ending_visarga)
        # Tags will go here as
        self.tags = []

    def setMorphologicalTags(self, t):
        """ Set Morphological Tags on Sanskrit Object

            Params:
               t (list): List of morphological tags
        """
        self.tags.extend(t)
        return self.tags

    def getMorphologicalTags(self):
        """ Morphological Tags on object """
        return self.tags


@contextmanager
def outputctx(strict_io):
    global denormalize
    save_denormalize = denormalize
    denormalize = not strict_io
    yield
    denormalize = save_denormalize


if __name__ == "__main__":
    import argparse

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = argparse.ArgumentParser(description='SanskritObject')
        # String to encode
        parser.add_argument('data', nargs="?", type=str, default="idam adbhutam")
        # Input Encoding (autodetect by default)
        parser.add_argument('--input-encoding', type=str, default=None)
        # Output Encoding (Devanagari by default)
        parser.add_argument('--output-encoding', type=str, default="Devanagari")

        return parser.parse_args()

    def main():
        args = getArgs()
        print(args.data)
        if args.input_encoding is None:
            ie = None
        else:
            ie = SCHEMES[args.input_encoding]

        oe = SCHEMES[args.output_encoding]

        s = SanskritObject(args.data, ie)
        print(s.transcoded(oe))

    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Intro
======
Get varnas in a pratyahara:

.. code:: python

    >>> from sanskrit_parser.base.maheshvara_sutra import MaheshvaraSutras
    >>> MS = MaheshvaraSutras()
    >>> jaS = SanskritImmutableString('jaS', encoding=SLP1)
    >>> print(MS.getPratyahara(jaS))
    jabagaqada

Check if a varna is in a pratyahara:

.. code:: python

    >>> g = SanskritImmutableString('g')
    >>> print(MS.isInPratyahara(jaS, g))
    True
    >>> k = SanskritImmutableString('k')
    >>> print(MS.isInPratyahara(jaS, k))
    False



Command line usage
==================

::

    $ python -m sanskrit_parser.base.maheshvara_sutra --encoding SLP1 --pratyahara jaS
    aiuR fxk eoN EOc hayavaraw laR YamaNaRanam JaBaY GaQaDaz jabagaqadaS KaPaCaWaTacawatav kapay Sazasar hal
    जश्
    जबगडद

"""

from __future__ import print_function
from . import sanskrit_base
import re
import six


class MaheshvaraSutras(object):
    """
    Singleton MaheshvaraSutras class
    Attributes:
    MS(SanskritImmutableString) : Internal representation of mAheshvara sutras
    MSS(str)           : Canonical (SLP1) representation
    """

    def __init__(self):
        """
        Initialize Maheshvara Sutras object
        """
        # Note that a space is deliberately left after each it to help in
        # demarcating them.
        self.MS = sanskrit_base.SanskritImmutableString(
            u'अइउण् ऋऌक् एओङ् ऐऔच् हयवरट् लण् ञमङणनम् झभञ् घढधष् जबगडदश् खफछठथचटतव् कपय् शषसर् हल् ',
            sanskrit_base.DEVANAGARI)
        # SLP1 version for internal operations
        self.MSS = self.MS.canonical()

    def __str__(self):
        # Use SLP1 for default string output
        return self.MSS

    def getPratyahara(self, p, longp=True, remove_a=False, dirghas=False):
        """
        Return list of varnas covered by a pratyahara

        Args:
              p(:class:SanskritImmutableString): Pratyahara
              longp(boolean :optional:): When True (default), uses long pratyaharas
              remove_a(boolean :optional:): When True, removes intermediate 'a'.This is better for computational use
              dirghas(boolean :optional:) When True (default=False) adds dirgha vowels to the returned varnas
        Returns:
              (SanskritImmutableString): List of varnas to the same encoding as p
        """

        # SLP1 encoded pratyahara string
        ps = p.canonical()
        # it - halantyam
        pit = ps[-1]
        # Non it - all except it
        pnit = ps[:-1]
        # Non it position
        pnpos = self.MSS.find(pnit)
        # It position - space added to match it marker in internal
        # representation
        if longp:  # Find last occurence of it
            pitpos = self.MSS.rfind(pit + ' ', pnpos)
        else:  # Find first occurence of it
            pitpos = self.MSS.find(pit + ' ', pnpos)
        # Substring. This includes intermediate its and spaces
        ts = self.MSS[pnpos:pitpos]
        # Replace its and spaces
        ts = re.sub('. ', '', ts)
        # Remove अकारः मुखसुखार्थः
        if remove_a:
            ts = ts[0] + ts[1:].replace('a', '')
        # Add dIrgha vowels if requested
        if dirghas:
            ts = ts.replace('a', 'aA').replace('i', 'iI').replace('u', 'uU').replace('f', 'fF').replace('x', 'xX')
        return sanskrit_base.SanskritImmutableString(ts, sanskrit_base.SLP1)

    def isInPratyahara(self, p, v, longp=True):
        """
        Checks whether a given varna is in a pratyahara

        Args:
            p(SanskritImmutableString): Pratyahara
            v(SanskritImmutableString): Varna
            longp(boolean :optional:): When True (default), uses long pratyaharas
        Returns
             boolean: Is v in p?
        """

        vs = v.canonical()
        # १ . १ . ६९ अणुदित् सवर्णस्य चाप्रत्ययः
        # So, we change long and pluta vowels to short ones in the input string
        # Replace long vowels with short ones (note SLP1 encoding)
        vs = re.sub('[AIUFX]+', lambda m: m.group(0).lower(), vs)
        # Remove pluta
        vs = vs.replace('3', '')

        # Convert Pratyahara into String
        # the 'a' varna needs special treatment - we remove the
        # अकारः मुखसुखार्थः before searching!
        pos = self.getPratyahara(p, longp, remove_a=vs[0] == 'a').canonical()

        # Check if varna String is in Pratyahara String
        return (pos.find(vs) != -1)


if __name__ == "__main__":
    import argparse

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = argparse.ArgumentParser(description='SanskritImmutableString')
        # Pratyahara - print out the list of varnas in this
        parser.add_argument('--pratyahara', type=str, default="ik")
        # Varna. Optional. Check if this varna is in pratyahara above
        parser.add_argument('--varna', type=str, default=None)
        # Encoding Optional
        parser.add_argument('--encoding', type=str, default=None)
        # Short pratyaharas
        parser.add_argument('--short', action='store_true')
        # Remove intermediate as
        parser.add_argument('--remove-a', action='store_true')
        # Include dIrghas when returning the pratyAhAra
        parser.add_argument('--dirghas', action='store_true', default=False)

        return parser.parse_args()

    def main():
        args = getArgs()
        m = MaheshvaraSutras()
        print(m)
        if args.encoding is not None:
            e = sanskrit_base.SCHEMES[args.encoding]
        else:
            e = None
        p = sanskrit_base.SanskritImmutableString(args.pratyahara, e)
        longp = not args.short
        print(six.text_type(p.devanagari()))
        print(six.text_type(m.getPratyahara(p, longp, args.remove_a, args.dirghas).devanagari()))
        if args.varna is not None:
            v = sanskrit_base.SanskritImmutableString(args.varna, e)
            print(u"Is {} in {}?".format(v.devanagari(),
                                         p.devanagari()))
            print(m.isInPratyahara(p, v, longp))

    main()



# -*- coding: utf-8 -*-

"""
Intro
=====
Provide interface with the INRIA XML database released
by Prof. Gerard Huet
https://gitlab.inria.fr/huet/Heritage_Resources
(Thanks Prof. Huet for making the data available!)

Inspired by https://github.com/drdhaval2785/inriaxmlwrapper
(Thanks @drdhaval2785!)

@author: Avinash Varna (@avinashvarna)

Usage
=====
The InriaXMLWrapper utility class can also be used to lookup tags:

.. code:: python

    >>> from sanskrit_parser.util.inriaxmlwrapper import InriaXMLWrapper
    >>> db = InriaXMLWrapper()
    >>> db_tags = db.get_tags('hares')
    >>> tags == db_tags
    True


Command line usage
==================

::

    $ python -m sanskrit_parser.util.inriaxmlwrapper hares
    INFO:root:Pickle file found, loading at 2017-07-31 14:35:56.093000
    INFO:root:Loading finished at 2017-07-31 14:35:59.159000, took 3.066000 s
    INFO:root:Cached 666994 forms for fast lookup
    Getting tags for hares
    ('hf#1', set(['cj', 'snd', 'prim', 'para', 'md', 'sys', 'prs', 'v', 'np', 'sg', 'op']))
    ('hari#1', set(['na', 'mas', 'sg', 'gen']))
    ('hari#1', set(['na', 'mas', 'abl', 'sg']))
    ('hari#1', set(['na', 'fem', 'sg', 'gen']))
    ('hari#1', set(['na', 'fem', 'abl', 'sg']))
    ('hari#2', set(['na', 'mas', 'sg', 'gen']))
    ('hari#2', set(['na', 'mas', 'abl', 'sg']))
    ('hari#2', set(['na', 'fem', 'sg', 'gen']))
    ('hari#2', set(['na', 'fem', 'abl', 'sg']))

"""

from __future__ import print_function
import requests
import os
import shutil
import inspect
from lxml import etree
from collections import defaultdict
from io import BytesIO
import logging
import time
import datetime
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SCHEMES
from sanskrit_parser.util.lexical_lookup import LexicalLookup
from sanskrit_parser.util.inriatagmapper import inriaTagMapper

try:
    import cPickle as pickle
except ImportError:
    import pickle


class InriaXMLWrapper(LexicalLookup):
    """
    Class to interface with the INRIA XML database released
    by Prof. Gerard Huet
    https://gitlab.inria.fr/huet/Heritage_Resources
    """
    base_url = "https://raw.githubusercontent.com/drdhaval2785/inriaxmlwrapper/master/"
    xml_files = ["roots", "nouns", "adverbs", "final", "parts", "pronouns", "upasargas", "all"]
    old_base_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    folder = "data"
    old_dir = os.path.join(old_base_dir, folder)

    def __init__(self, files_list=['all'], logger=None):
        for f in files_list:
            if f not in self.xml_files:
                raise Exception(f + "is not a valid file name")
        self.files = files_list
        self.files.sort()
        self.pickle_file = "_".join(self.files) + ".pickle"
        if ["all"] == self.files:
            self.files = self.xml_files[:-1]
            self.pickle_file = "_all.pickle"
        self.logger = logger or logging.getLogger(__name__)
        self._relocate_data()
        self._load_forms()

    @staticmethod
    def _relocate_data():
        if os.path.exists(InriaXMLWrapper.old_dir):
            shutil.move(InriaXMLWrapper.old_dir, InriaXMLWrapper.base_dir)

    def _get_files(self):
        """ Download files if not present in cache """
        if not os.path.exists(self.base_dir):
            self.logger.debug("Data cache not found. Creating.")
            os.makedirs(self.base_dir)
        for f in self.files:
            filename = "SL_" + f + ".xml"
            if not os.path.exists(os.path.join(self.base_dir, filename)):
                self.logger.debug("%s not found. Downloading it", filename)
                r = requests.get(self.base_url + filename, stream=True)
                with open(os.path.join(self.base_dir, filename), "wb") as fd:
                    for chunk in r.iter_content(chunk_size=128):
                        fd.write(chunk)

    def _generate_dict(self):
        """ Create dict with mapping
            form : set([<tag 1>, ... , <tag n>])
            and pickle it
        """
        self._get_files()
        self.logger.debug("Parsing files into dict for faster lookup")
        self.forms = defaultdict(list)
        for f in self.files:
            filename = "SL_" + f + ".xml"
            self.logger.debug("Parsing %s", filename)
            tree = etree.parse(os.path.join(self.base_dir, filename))
            for elem in tree.iterfind('f'):
                form = elem.get('form')
                self.forms[form].append(etree.tostring(elem).strip())
        self.logger.debug("Pickling forms database for faster loads")
        with open(os.path.join(self.base_dir, self.pickle_file), "wb") as fd:
            pickle.dump(self.forms, fd, pickle.HIGHEST_PROTOCOL)

    def _load_forms(self):
        """ Load/create dict of tags for forms """
        pickle_path = os.path.join(self.base_dir, self.pickle_file)
        if os.path.exists(pickle_path):
            self.logger.debug("Pickle file found, loading at %s", datetime.datetime.now())
            start = time.time()
            with open(pickle_path, "rb") as fd:
                self.forms = pickle.load(fd)
            self.logger.debug("Loading finished at %s, took %f s",
                              datetime.datetime.now(),
                              time.time() - start)
        else:
            self.logger.debug("Pickle file not found, creating ...")
            self._generate_dict()
        self.logger.debug("Cached %d forms for fast lookup", len(self.forms))

    def _xml_to_tags(self, word):
        # FIXME - This is currently from sanskritmark. Check if this can be simplified
        if word in self.forms:
            tags = self.forms[word]
            results = []
            for tag in tags:
                root = etree.parse(BytesIO(tag)).getroot()
                # The next two steps require explanation. In Gerard's XML files,
                # All possible attributes are given as children of 'f'. The last
                # child is always 's' which stores the stem. All other children
                # are the various possible word attributes. Given as 'na' or 'v'
                # etc. Gio
                children = root.getchildren()[:-1]  # attributes
                baseword = root.getchildren()[-1].get('stem').strip()  # 's' stem
                attributes = []
                for child in children:
                    taglist = child.xpath(
                        './/*')  # Fetches all elements (abbreviations) of a particular verb / word characteristics.
                    output = [child.tag]  # The first member of output list is the tag of element 'v', 'na' etc.
                    output = output + [tagitem.tag for tagitem in
                                       taglist]  # Other tags (abbreviations) and add it to output list.
                    attributes.append(output)
                for attrib in attributes:
                    results.append((baseword, set(attrib)))
            return results
        else:
            return None

    def valid(self, word):
        return word in self.forms

    def get_tags(self, word, tmap=True):
        tags = self._xml_to_tags(word)
        if tmap and (tags is not None):
            tags = inriaTagMapper(tags)
        return tags


if __name__ == "__main__":
    from argparse import ArgumentParser

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = ArgumentParser(description='Interface to INRIA XML database')
        # Input Encoding (autodetect by default)
        parser.add_argument('--input-encoding', type=str, default=None)
        parser.add_argument('--loglevel', type=str, default="info",
                            help="logging level. Can be any level supported by logging module")

        parser.add_argument('word', nargs='?', type=str,
                            default=None,
                            help="Word to look up")
        parser.add_argument('--no-map-tags', dest='map_tags',
                            action='store_false')
        return parser.parse_args()

    def main():
        args = getArgs()
        if args.input_encoding is None:
            ie = None
        else:
            ie = SCHEMES[args.input_encoding]

        if args.loglevel:
            numeric_level = getattr(logging, args.loglevel.upper(), None)
            if not isinstance(numeric_level, int):
                raise ValueError('Invalid log level: %s' % args.loglevel)
            logging.basicConfig(level=numeric_level)

        word_in = SanskritImmutableString(args.word, encoding=ie).canonical()
        xmlDB = InriaXMLWrapper()
        print("Getting tags for", word_in)
        tags = xmlDB.get_tags(word_in, tmap=args.map_tags)
        if tags is not None:
            for t in tags:
                print(t)

    main()



from flask import Blueprint, redirect
import flask_restplus
from flask_restplus import Resource
from random import randint
import subprocess
from os import path

from sanskrit_parser.base.sanskrit_base import SanskritObject, SLP1
from sanskrit_parser.parser.sandhi_analyzer import LexicalSandhiAnalyzer
from sanskrit_parser.parser.datastructures import VakyaGraph

URL_PREFIX = '/v1'
api_blueprint = Blueprint(
    'sanskrit_parser', __name__,
    template_folder='templates'
)

api = flask_restplus.Api(app=api_blueprint, version='1.0', title='sanskrit_parser API',
                         description='For detailed intro and to report issues: see <a href="https://github.com/kmadathil/sanskrit_parser">here</a>. '
                                     'A list of REST and non-REST API routes avalilable on this server: <a href="../sitemap">sitemap</a>.',
                         default_label=api_blueprint.name,
                         prefix=URL_PREFIX, doc='/docs')

analyzer = LexicalSandhiAnalyzer()


def jedge(pred, node, label):
    return (node.pada.devanagari(strict_io=False),
            jtag(node.getMorphologicalTags()),
            SanskritObject(label, encoding=SLP1).devanagari(strict_io=False),
            pred.pada.devanagari(strict_io=False))


def jnode(node):
    """ Helper to translate parse node into serializable format"""
    return (node.pada.devanagari(strict_io=False),
            jtag(node.getMorphologicalTags()), "", "")


def jtag(tag):
    """ Helper to translate tag to serializable format"""
    return (tag[0].devanagari(strict_io=False), [t.devanagari(strict_io=False) for t in list(tag[1])])


def jtags(tags):
    """ Helper to translate tags to serializable format"""
    return [jtag(x) for x in tags]


@api.route('/tags/<string:p>')
class Tags(Resource):
    def get(self, p):
        """ Get lexical tags for p """
        pobj = SanskritObject(p, strict_io=False)
        tags = analyzer.getMorphologicalTags(pobj)
        if tags is not None:
            ptags = jtags(tags)
        else:
            ptags = []
        r = {"input": p, "devanagari": pobj.devanagari(), "tags": ptags}
        return r


@api.route('/splits/<string:v>')
class Splits(Resource):
    def get(self, v):
        """ Get lexical tags for v """
        vobj = SanskritObject(v, strict_io=False, replace_ending_visarga=None)
        g = analyzer.getSandhiSplits(vobj)
        if g:
            splits = g.find_all_paths(10)
            jsplits = [[ss.devanagari(strict_io=False) for ss in s] for s in splits]
        else:
            jsplits = []
        r = {"input": v, "devanagari": vobj.devanagari(), "splits": jsplits}
        return r


@api.route('/analyses/<string:v>')
class Morpho(Resource):
    def get(self, v):
        """ Get morphological tags for v """
        vobj = SanskritObject(v, strict_io=False, replace_ending_visarga=None)
        g = analyzer.getSandhiSplits(vobj, tag=True)
        if g:
            splits = g.find_all_paths(10, score=True)
        else:
            splits = []
        mres = {}
        plotbase = {}
        for sp in splits:
            bn = f"{randint(0,9999):4}"
            vg = VakyaGraph(sp, max_parse_dc=5)
            sl = "_".join([n.devanagari(strict_io=False)
                           for n in sp])
            for (ix, p) in enumerate(vg.parses):
                if sl not in mres:
                    mres[sl] = []
                t = []
                for n in p:
                    preds = list(p.predecessors(n))
                    if preds:
                        pred = preds[0]  # Only one
                        lbl = p.edges[pred, n]['label']
                        t.append(jedge(pred, n, lbl))
                    else:
                        t.append(jnode(n))
                mres[sl].append(t)
            plotbase[sl] = bn
            try:
                vg.write_dot(f"static/{bn}.dot")
            except Exception:
                pass
        r = {"input": v, "devanagari": vobj.devanagari(), "analysis": mres, "plotbase": plotbase}
        return r

    @api.route('/graph/<string:v>')
    class Graph(Resource):
        def get(self, v):
            """ Get graph for v """
            if not path.exists(f"static/{v}.dot.png"):
                subprocess.run(f"dot -Tpng static/{v}.dot -O", shell=True)
            return redirect(f"/static/{v}.dot.png")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Intro
=====

Sandhi Analyzer for Sanskrit words

@author: Karthik Madathil (github: @kmadathil)

Usage
=====

Use the ``LexicalSandhiAnalyzer`` to split a sentence (wrapped in a
``SanskritObject``) and retrieve the top 10 splits:

.. code:: python

    >>> from __future__ import print_function
    >>> from sanskrit_parser.parser.sandhi_analyzer import LexicalSandhiAnalyzer
    >>> from sanskrit_parser.base.sanskrit_base import SanskritObject, SLP1
    >>> sentence = SanskritObject("astyuttarasyAMdishidevatAtmA")
    >>> analyzer = LexicalSandhiAnalyzer()
    >>> splits = analyzer.getSandhiSplits(sentence).findAllPaths(10)
    >>> for split in splits:
    ...    print(split)
    ...
    [u'asti', u'uttarasyAm', u'diSi', u'devatA', u'AtmA']
    [u'asti', u'uttarasyAm', u'diSi', u'devat', u'AtmA']
    [u'asti', u'uttarasyAm', u'diSi', u'devata', u'AtmA']
    [u'asti', u'uttara', u'syAm', u'diSi', u'devatA', u'AtmA']
    [u'asti', u'uttarasyAm', u'diSi', u'devatA', u'at', u'mA']
    [u'asti', u'uttarasyAm', u'diSi', u'de', u'vatA', u'AtmA']
    [u'asti', u'uttarasyAm', u'diSi', u'devata', u'at', u'mA']
    [u'asti', u'uttas', u'rasyAm', u'diSi', u'devat', u'AtmA']
    [u'asti', u'uttara', u'syAm', u'diSi', u'devat', u'AtmA']
    [u'asti', u'uttarasyAm', u'diSi', u'de', u'avatA', u'AtmA']

The sandhi_analyzer can also be used to look up the tags for a given
word form: (Note that the database stores words
ending in visarga with an 's' at the end)

.. code:: python

    >>> word = SanskritObject('hares')
    >>> tags = analyzer.getMorphologicalTags(word)
    >>> for tag in tags:
    ...    print(tag)
    ...
    ('hf#1', set(['cj', 'snd', 'prim', 'para', 'md', 'sys', 'prs', 'v', 'np', 'sg', 'op']))
    ('hari#1', set(['na', 'mas', 'sg', 'gen']))
    ('hari#1', set(['na', 'mas', 'abl', 'sg']))
    ('hari#1', set(['na', 'fem', 'sg', 'gen']))
    ('hari#1', set(['na', 'fem', 'abl', 'sg']))
    ('hari#2', set(['na', 'mas', 'sg', 'gen']))
    ('hari#2', set(['na', 'mas', 'abl', 'sg']))
    ('hari#2', set(['na', 'fem', 'sg', 'gen']))
    ('hari#2', set(['na', 'fem', 'abl', 'sg']))



"""

from __future__ import print_function
from sanskrit_parser.util.lexical_lookup_factory import LexicalLookupFactory
import sanskrit_parser.base.sanskrit_base as SanskritBase

from .sandhi import Sandhi
import logging
from .datastructures import SandhiGraph
from argparse import ArgumentParser


try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

logger = logging.getLogger(__name__)


class LexicalSandhiAnalyzer(object):
    """ Singleton class to hold methods for Sanskrit lexical sandhi analysis.

        We define lexical sandhi analysis to be the process of taking an input sequence
        and transforming it to a collection (represented by a DAG) of potential sandhi
        splits of the sequence. Each member of a split is guaranteed to be a valid
        lexical form.
    """

    sandhi = Sandhi()  # Singleton!

    def __init__(self, lexical_lookup="combined"):
        forms = LexicalLookupFactory.create(lexical_lookup)
        self.forms = forms
        pass

    def getMorphologicalTags(self, obj, tmap=True):
        """ Get Morphological tags for a word

            Params:
                obj(SanskritString): word
                tmap(Boolean=True): If True, maps
                    tags to our format
            Returns
                list: List of (base, tagset) pairs
        """
        ot = obj.canonical()
        tags = self.forms.get_tags(ot, tmap)
        return tags

    def hasTag(self, obj, name, tagset):
        """ Check if word matches morhphological tags

            Params:
                obj(SanskritString): word
                name(str): name in tag
                tagset(set): set of tag elements
            Returns
                list: List of (base, tagset) pairs for obj that
                      match (name,tagset), or None
        """
        morphological_tags = self.getMorphologicalTags(obj)
        if morphological_tags is None:
            return None
        assert (name is not None) or (tagset is not None)
        r = []
        for li in morphological_tags:
            # Name is none, or name matches
            # Tagset is None, or all its elements are found in tagset
            if ((name is None) or name.canonical() == li[0]) and \
                    ((tagset is None) or tagset.issubset(li[1])):
                r.append(li)
        if r == []:
            return None
        else:
            return r

    def tagSandhiGraph(self, g):
        ''' Tag a Sandhi Graph with morphological tags for each node

         Params:
            g (SandhiGraph) : input lexical sandhi graph
        '''
        for n in g:
            # Avoid start and end
            if isinstance(n, SanskritBase.SanskritObject):
                t = self.getMorphologicalTags(n)
                logger.debug("Got tags %s for %s", t, n)
                n.setMorphologicalTags(t)

    def getSandhiSplits(self, o, tag=False):
        ''' Get all valid Sandhi splits for a string

            Params:
              o(SanskritString): Input object
              tag(Boolean)     : When True (def=False), return a
                                 morphologically tagged graph
            Returns:
              SandhiGraph : DAG all possible splits
        '''
        self.dynamic_scoreboard = {}
        # Transform to internal canonical form
        s = o.canonical()
        # Initialize an empty graph to hold the splits
        self.splits = SandhiGraph()
        # _possible_splits updates graph in self.splits with nodes and returns roots
        roots = self._possible_splits(s)
        if tag and len(roots) > 0:
            self.tagSandhiGraph(self.splits)
        if len(roots) == 0:
            return None
        else:
            self.splits.add_roots(roots)
            return self.splits

    def _possible_splits(self, s):
        ''' private method to dynamically compute all sandhi splits

            Used by getSandhiSplits
            Adds the individual splits to the graph self.splits and returns
            the roots of the subgraph corresponding to the split of s
           Params:
              s(string): Input SLP1 encoded string
            Returns:
              roots : set of roots of subgraph corresponding to possible splits of s
        '''
        logger.debug("Splitting " + s)

        @lru_cache(256)
        def _is_valid_word(ss):
            r = self.forms.valid(ss)
            return r

        def _sandhi_splits_all(s, start=None, stop=None):
            obj = SanskritBase.SanskritImmutableString(s, encoding=SanskritBase.SLP1)
            splits = self.sandhi.split_all(obj, start, stop)
            return splits

        roots = set()

        # Memoization for dynamic programming - remember substrings that've
        # been seen before
        if s in self.dynamic_scoreboard:
            logger.debug("Found {} in scoreboard".format(s))
            return self.dynamic_scoreboard[s]

        # If a space is found in a string, stop at that space
        spos = s.find(" ")
        stop = None if spos == -1 else spos

        s_c_list = _sandhi_splits_all(s, start=0, stop=stop)
        logger.debug("s_c_list: " + str(s_c_list))
        if s_c_list is None:
            s_c_list = []

        node_cache = {}

        for (s_c_left, s_c_right) in s_c_list:
            # Is the left side a valid word?
            if _is_valid_word(s_c_left):
                logger.debug("Valid left word: " + s_c_left)
                # For each split with a valid left part, check it there are
                # valid splits of the right part
                if s_c_right and s_c_right != '':
                    logger.debug("Trying to split:" + s_c_right)
                    r_roots = self._possible_splits(s_c_right.strip())
                    # if there are valid splits of the right side
                    if r_roots:
                        # Make sure we got a set of roots back
                        assert isinstance(r_roots, set)
                        # if there are valid splits of the right side
                        if s_c_left not in node_cache:
                            # Extend splits list with s_c_left appended with
                            # possible splits of s_c_right
                            t = SanskritBase.SanskritObject(s_c_left, encoding=SanskritBase.SLP1)
                            node_cache[s_c_left] = t
                        else:
                            t = node_cache[s_c_left]
                        roots.add(t)
                        if not self.splits.has_node(t):
                            self.splits.add_node(t)
                        self.splits.append_to_node(t, r_roots)
                else:  # Null right part
                    # Why cache s_c_left here? To handle the case
                    # where the same s_c_left appears with a null and non-null
                    # right side.
                    if s_c_left not in node_cache:
                        t = SanskritBase.SanskritObject(s_c_left, encoding=SanskritBase.SLP1)
                        node_cache[s_c_left] = t
                    else:
                        t = node_cache[s_c_left]
                    # Extend splits list with s_c_left appended with
                    # possible splits of s_c_right
                    roots.add(t)
                    if not self.splits.has_node(t):
                        self.splits.add_node(t)
                    self.splits.add_end_edge(t)
            else:
                logger.debug("Invalid left word: " + s_c_left)
        # Update scoreboard for this substring, so we don't have to split
        # again
        self.dynamic_scoreboard[s] = roots
        if len(roots) == 0:
            logger.debug("No splits found, returning empty set")
        else:
            logger.debug("Roots: %s", roots)
        return roots


def getArgs(argv=None):
    """
      Argparse routine.
      Returns args variable
    """
    # Parser Setup
    parser = ArgumentParser(description='Lexical Sandhi Analyzer')
    # String to encode
    parser.add_argument('data', nargs="?", type=str, default="adhi")
    # Input Encoding (autodetect by default)
    parser.add_argument('--input-encoding', type=str, default=None)
    # Filter by base name
    parser.add_argument('--base', type=str, default=None)
    # Filter by tag set
    parser.add_argument('--tag-set', type=str, default=None, nargs="+")
    parser.add_argument('--tags', dest='split', action='store_false')
    parser.add_argument('--max-paths', type=int, default=10)
    parser.add_argument('--lexical-lookup', type=str, default="combined")
    parser.add_argument('--strict-io', action='store_true',
                        help="Do not modify the input/output string to match conventions", default=False)
    parser.add_argument('--no-score', dest="score", action='store_false',
                        help="Don't use the lexical scorer to score the splits and reorder them")
    parser.add_argument('--no-map-tags', dest='map_tags',
                        action='store_false', help="show raw (unmapped to our standard set) tags")
    return parser.parse_args(argv)


def main(argv=None):
    graph = None
    args = getArgs(argv)
    if args.strict_io:
        print("Interpreting input strictly")
    else:
        print("Interpreting input loosely (strict_io set to false)")
    print("Input String:", args.data)

    s = LexicalSandhiAnalyzer(args.lexical_lookup)
    if args.input_encoding is None:
        ie = None
    else:
        ie = SanskritBase.SCHEMES[args.input_encoding]
    with SanskritBase.outputctx(args.strict_io):
        if not args.split:
            i = SanskritBase.SanskritNormalizedString(args.data, encoding=ie,
                                                      strict_io=args.strict_io,
                                                      replace_ending_visarga='s')
            print("Input String in SLP1:", i.canonical())
            ts = s.getMorphologicalTags(i, tmap=args.map_tags)
            print("Morphological tags:")
            if ts is not None:
                for t in ts:
                    print(t)
            # Possible rakaranta
            # Try by replacing end visarga with 'r' instead
            elif not args.strict_io:
                i = SanskritBase.SanskritNormalizedString(args.data, encoding=ie,
                                                          strict_io=args.strict_io,
                                                          replace_ending_visarga='r')
                ts = s.getMorphologicalTags(i)
                if ts is not None:
                    print("Input String in SLP1:", i.canonical())
                    for t in ts:
                        print(t)
            if args.tag_set or args.base:
                if args.tag_set is not None:
                    g = set(args.tag_set)
                else:
                    g = None
                if args.base is not None:
                    b = SanskritBase.SanskritNormalizedString(args.base)
                else:
                    b = None
                print(s.hasTag(i, b, g))
        else:
            import time
            i = SanskritBase.SanskritNormalizedString(args.data, encoding=ie,
                                                      strict_io=args.strict_io,
                                                      replace_ending_visarga=None)
            print("Input String in SLP1:", i.canonical())
            print("Start Split")
            start_split = time.time()
            graph = s.getSandhiSplits(i)
            end_graph = time.time()
            print("End DAG generation")
            if graph:
                logger.debug("Graph has %d nodes and %d edges" % (len(graph.G.nodes()), len(graph.G.edges())))
                splits = graph.find_all_paths(max_paths=args.max_paths, score=args.score)
                print("End pathfinding", time.time())
                print("Splits:")
                if splits:
                    for split in splits:
                        print(split)
                else:
                    print("None")
            else:
                print("No Valid Splits Found")
            end_split = time.time()
            print("-----------")
            print("Performance")
            print("Time for graph generation = {0:0.6f}s".format(end_graph - start_split))
            print("Total time for graph generation + find paths = {0:0.6f}s".format(end_split - start_split))
        return graph


if __name__ == "__main__":
    main()



'''
Scorer for lexical splits

Uses word2vec based scoring the splits. The splits are
passed through sentencepiece first to handle possibly unseen words

Requires sentencepiece to be installed

@author: avinashvarna
'''

from __future__ import print_function, unicode_literals
import os
import sys
import six
import gensim
import logging
import requests
import sentencepiece as spm


class Scorer(object):

    # TODO - Install the data in a different way
    sentencepiece_file_url = "https://github.com/kmadathil/sanskrit_parser/blob/sanskrit_util/data/sentencepiece.model?raw=true"
    word2vec_file_url = "https://github.com/kmadathil/sanskrit_parser/blob/sanskrit_util/data/word2vec_model.dat?raw=true"
    base_dir = os.path.expanduser("~/.sanskrit_parser/data/")
    sentencepiece_file = os.path.join(base_dir, "sentencepiece.model")
    word2vec_file = os.path.join(base_dir, "word2vec_model.dat")

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self._get_file(self.sentencepiece_file, self.sentencepiece_file_url)
        self._get_file(self.word2vec_file, self.word2vec_file_url)
        self.sp = spm.SentencePieceProcessor()
        self.sp.Load(self.sentencepiece_file)
        self.model = gensim.models.Word2Vec.load(self.word2vec_file)

    def score_splits(self, splits):
        sentences = [" ".join(map(six.text_type, split)) for split in splits]
        return self.score_strings(sentences)

    def score_strings(self, sentences):
        self.logger.debug("Sentence = %s", sentences)
        pieces = [self.sp.EncodeAsPieces(sentence) for sentence in sentences]
        self.logger.debug("Pieces = %s", pieces)
        scores = self.model.score(pieces, total_sentences=len(sentences))
        self.logger.debug("Score = %s", scores)
        return scores

    def _get_file(self, local, url):
        if not os.path.exists(local):
            self.logger.debug("%s not found. Fetching from %s", local, url)
            r = requests.get(url, stream=True)
            with open(local, "wb") as fd:
                for chunk in r.iter_content(chunk_size=128):
                    fd.write(chunk)


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser(description='Lexical Scorer')
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('data', nargs="?", type=six.text_type, default=None)

    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    else:
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    score = Scorer().score_strings([args.data])
    print("Input:", args.data)
    print("Score =", score)



# -*- coding: utf-8 -*-
'''
Wrapper around data from Sanskrit data project

@author: avinashvarna
'''

from __future__ import print_function
import logging
import os
import sanskrit_util.analyze
import sanskrit_util.context
from sanskrit_util.schema import Nominal, Indeclinable, Verb, Gerund, Infinitive, ParticipleStem
from sanskrit_parser.util.lexical_lookup import LexicalLookup
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, DEVANAGARI, SLP1
import requests


class SanskritDataWrapper(LexicalLookup):

    # TODO - Install the data in a different way
    file_url = "https://github.com/kmadathil/sanskrit_parser/blob/sanskrit_util/data/sanskrit_data.db?raw=true"
    db_file = os.path.join(LexicalLookup.base_dir, 'sanskrit_data.db')

    def __init__(self, logger=None):
        self._get_file()
        config = {
            "DATABASE_URI": 'sqlite:///' + self.db_file,
            "DATA_PATH": ""}
        ctx = sanskrit_util.context.Context(config)
        self.analyzer = sanskrit_util.analyze.SimpleAnalyzer(ctx)
        self.tag_cache = {}
        self.logger = logger or logging.getLogger(__name__)

    def valid(self, word):
        self.logger.debug("Checking if %s is valid", word)
        if word in self.tag_cache:
            self.logger.debug("Cache hit")
            return True
        else:
            tags = self.get_tags(word, tmap=False)
            if tags is not None:
                self.logger.debug("Found tags")
                return True
        self.logger.debug("Returning False")
        return False

    def get_tags(self, word, tmap=True):
        self.logger.debug("Looking up tags for %s", word)
        if word in self.tag_cache:
            self.logger.debug("Cache hit")
            tags = self.tag_cache[word]
            self.logger.debug("Found tags %s", tags)
        else:
            tags = self.analyzer.analyze(word)
            if tags != []:
                self.logger.debug("Found tags %s", tags)
                self.tag_cache[word] = tags
        if tmap:
            tags = self.map_tags(tags)
            self.logger.debug("Found tags %s", tags)
        if tags != []:
            return tags
        else:
            return None

    def _get_file(self):
        if not os.path.exists(self.db_file):
            r = requests.get(self.file_url, stream=True)
            with open(os.path.join(self.db_file), "wb") as fd:
                for chunk in r.iter_content(chunk_size=128):
                    fd.write(chunk)

    vacanam = ['एकवचनम्', 'द्विवचनम्', 'बहुवचनम्']
    lingam = ['पुंल्लिङ्गम्', 'स्त्रीलिङ्गम्', 'नपुंसकलिङ्गम्', 'त्रिलिङ्गम्']
    vibhakti = ['प्रथमाविभक्तिः', 'द्वितीयाविभक्तिः', 'तृतीयाविभक्तिः',
                'चतुर्थीविभक्तिः', 'पञ्चमीविभक्तिः', 'षष्ठीविभक्तिः',
                'सप्तमीविभक्तिः', 'संबोधनविभक्तिः']
    lakAra = ['लट्', 'लुङ्', 'लङ्', 'लिट्', 'लृट्', 'लुट्', 'लृङ्', 'विधिलिङ्', 'लोट्', 'आशीर्लिङ्', 'आगमाभावयुक्तलुङ्']
    pada_prayoga = ['परस्मैपदम्', 'आत्मनेपदम्', 'उभयपदम्', 'कर्तरि', 'कर्मणि']
    puruSha = ['उत्तमपुरुषः', 'मध्यमपुरुषः', 'प्रथमपुरुषः']
    kRdanta = {
        ('fut', 'para'): ['Satf', 'Bavizyat', 'kartari'],  # 'कर्तरिभविष्यत्कृदन्त-परस्मैपदी',
        ('fut', 'atma'): ['SAnac', 'Bavizyat', 'kartari'],  # 'कर्तरिभविष्यत्कृदन्त-आत्मनेपदी',
        ('fut', 'pass'): ['kftya', 'Bavizyat', 'karmaRi'],  # 'कर्मणिभविष्यत्कृदन्तः',
        ('pres', 'para'): ['Satf', 'kartari'],  # 'कर्तरिवर्तमानकृदन्त-परस्मैपदी',
        ('pres', 'atma'): ['SAnac', 'kartari'],  # 'कर्तरिवर्तमानकृदन्त-आत्मनेपदी',
        ('pres', 'pass'): ['SAnac', 'karmaRi'],  # 'कर्मणिवर्तमानकृदन्तः',
        ('past', 'active'): ['ktavatu', 'kartari'],  # 'कर्तरिभूतकृदन्तः',
        ('past', 'pass'): ['kta', 'karmaRi'],  # 'कर्मणिभूतकृदन्तः',
        ('perf', 'para'): ['kvasu'],  # 'पूर्णभूतकृदन्त-परस्मैपदी',
        ('perf', 'atma'): ['kAnac'],  # 'पूर्णभूतकृदन्त-आत्मनेपदी'
    }

    def refresh(self, obj):
        if hasattr(obj, "id") and obj.id is not None:
            return self.analyzer.session.query(type(obj)).populate_existing().get(obj.id)
        else:
            return obj

    def map_nominal(self, obj):
        self.logger.debug("map_nominal %s", obj)
        tagset = set()
        obj = self.refresh(obj)
#         self.logger.debug("%s, gender_id = %d", obj, obj.gender_id)
        tagset.add(SanskritImmutableString(self.lingam[obj.gender_id - 1], DEVANAGARI))
        if obj.compounded:
            tagset.add(SanskritImmutableString("समासपूर्वपदनामपदम्", DEVANAGARI))
        else:
            tagset.add(SanskritImmutableString(self.vacanam[obj.number_id - 1], DEVANAGARI))
            tagset.add(SanskritImmutableString(self.vibhakti[obj.case_id - 1], DEVANAGARI))
        stem = obj.stem
        if type(stem) == ParticipleStem:
            mode = stem.mode.abbr
            voice = stem.voice.abbr
            for t in self.kRdanta[(mode, voice)]:
                tagset.add(SanskritImmutableString(t, SLP1))
        return (SanskritImmutableString(stem.name, SLP1), tagset)

    def map_verb(self, obj):
        tagset = set()
        newobj = self.refresh(obj)
        tagset.add(SanskritImmutableString(self.lakAra[newobj.mode.id - 1], DEVANAGARI))
        tagset.add(SanskritImmutableString(self.pada_prayoga[newobj.voice.id - 1], DEVANAGARI))
        tagset.add(SanskritImmutableString(self.puruSha[newobj.person.id - 1], DEVANAGARI))
        tagset.add(SanskritImmutableString(self.vacanam[newobj.number.id - 1], DEVANAGARI))
        return (SanskritImmutableString(newobj.root.name, SLP1), tagset)

    def map_tags(self, tags):
        out = []
        for t in tags:
            if type(t) == Nominal:
                otags = self.map_nominal(t)
                if otags not in out:
                    out.append(otags)
            elif type(t) == Verb:
                out.append(self.map_verb(t))
            elif type(t) == Indeclinable:
                out.append((SanskritImmutableString(t.name, SLP1),
                            set([SanskritImmutableString('avyayam', SLP1)])))
            elif type(t) == Gerund:
                newobj = self.refresh(t)
                out.append((SanskritImmutableString(newobj.root.name, SLP1),
                            set([SanskritImmutableString('ktvA', SLP1)])))
            elif type(t) == Infinitive:
                newobj = self.refresh(t)
                out.append((SanskritImmutableString(newobj.root.name, SLP1),
                            set([SanskritImmutableString('tumun', SLP1)])))
            else:
                out.append(t)
        return out


if __name__ == "__main__":
    args = LexicalLookup.getArgs()
    if args.loglevel:
        numeric_level = getattr(logging, args.loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % args.loglevel)
        logging.basicConfig(level=numeric_level)
    SanskritDataWrapper().main(args)



# -*- coding: utf-8 -*-
"""
Map tags generated by inriaxmlwrapper to a better set of names

@author: Karthik Madathil(@kmadathil)

"""

from __future__ import print_function
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SCHEMES

inriatags = [('prim', 'प्राथमिकः'),
             ('ca', 'णिजन्तः'),
             ('int', 'यङन्तः'),
             ('des', 'सन्नन्तः'),
             ('prs', 'kartari'),
             ('pas', 'karmaRi'),
             ('sys-md-pr', 'लट्'),
             ('sys-md-ip', 'लोट्'),
             ('sys-md-op', 'विधिलिङ्'),
             ('sys-md-im', 'लङ्'),
             ('sys-tp-fut', 'लृट्'),
             ('sys-tp-prf', 'लिट्'),
             ('sys-tp-aor', 'लुङ्'),
             ('sys-tp-inj', 'आगमाभावयुक्तलुङ्'),
             ('sys-tp-cnd', 'लृङ्'),
             ('sys-tp-ben', 'आशीर्लिङ्'),
             ('sys-pef', 'लुट्'),
             ('fst', 'उत्तमपुरुषः'),
             ('snd', 'मध्यमपुरुषः'),
             ('trd', 'प्रथमपुरुषः'),
             ('na-nom', 'प्रथमाविभक्तिः'),
             ('na-voc', 'संबोधनविभक्तिः'),
             ('na-acc', 'द्वितीयाविभक्तिः'),
             ('na-ins', 'तृतीयाविभक्तिः'),
             ('na-dat', 'चतुर्थीविभक्तिः'),
             ('na-abl', 'पञ्चमीविभक्तिः'),
             ('na-gen', 'षष्ठीविभक्तिः'),
             ('na-loc', 'सप्तमीविभक्तिः'),
             ('sg', 'एकवचनम्'),
             ('du', 'द्विवचनम्'),
             ('pl', 'बहुवचनम्'),
             ('mas', 'पुंल्लिङ्गम्'),
             ('fem', 'स्त्रीलिङ्गम्'),
             ('neu', 'नपुंसकलिङ्गम्'),
             ('dei', 'सङ्ख्या'),
             ('uf', 'अव्ययम्'),
             ('ind', 'क्रियाविशेषणम्'),
             ('interj', 'उद्गारः'),
             ('parti', 'निपातम्'),
             ('prep', 'karmapravacanIyaH'),
             ('conj', 'संयोजकः'),
             ('tasil', 'तसिल्'),
             ('vu', 'अव्ययधातुरूप'),
             ('vu', 'अव्यय'),
             ('iv-inf', 'तुमुन्'),
             ('iv-per', 'per'),
             ('ab', 'अव्ययधातुरूप'),
             ('ab', 'ktvA'),
             ('abs', 'lyap'),
             ('kr', 'kfdantaH'),
             ('ppp', 'karmaRi'),
             ('ppp', 'kta'),
             ('ppa', 'ktavatu'),
             ('ppa', 'kartari'),
             ('pprp', 'SAnac'),
             ('pprp', 'karmaRi'),
             ('ppr', 'kartari'),
             ('ppr-para', 'Satf'),
             ('ppr-atma', 'SAnac'),
             ('ppr', 'kartari'),
             ('ppft', 'kartari'),
             ('ppft-para', 'kvasu'),
             ('ppft-atma', 'kAnac'),
             ('pfutp', 'karmaRi'),
             ('pfutp', 'kftya'),
             ('pfut', 'kartari'),
             ('pfut', 'Bavizyat'),
             ('pfut-para', 'Satf'),
             ('pfut-atma', 'SanaC'),
             ('gya', 'य'),
             ('iya', 'ईय'),
             ('tav', 'तव्य'),
             ('para', 'परस्मैपदम्'),
             ('atma', 'आत्मनेपदम्'),
             ('pas', 'आत्मनेपदम्'),  # ('pa', 'कृदन्तः'),
             ('iic', 'समासपूर्वपदनामपदम्'),
             ('iip', 'समासपूर्वपदकृदन्तः'),
             ('iiv', 'समासपूर्वपदधातुः'),
             ('upsrg', 'उपसर्गः')]


def _inriaTagsToDb(tag):
    # ('tag','ourtag)
    itag = tag[0]
    otag = tag[1]
    if itag.find('-') != -1:
        iset = set(itag.split('-'))
    else:
        iset = set([itag])
    return (iset, SanskritImmutableString(otag, encoding=SCHEMES['Devanagari']))


inriatagdb = list(map(_inriaTagsToDb, inriatags))


def inriaMapTag(tag):
    ''' Map an INRIA tag to our format

    Params:
       tag(tuple) : (<stem>,set([inria tags]))
    Returns
       tuple : (<stem>,set([our tags]))
    '''
    stem = tag[0]
    tset = tag[1]
    olist = []
    for s in inriatagdb:
        if s[0] <= tset:
            olist.append(s[1])
    return (SanskritImmutableString(stem, SCHEMES['SLP1']), set(olist))


def inriaTagMapper(tags):
    ''' Map an INRIA tag to our format

    Params:
       tags: List of INRIA format tags
    Returns
       list[(<stem>,set([our tags])) ...]
    '''
    return list(map(inriaMapTag, tags))


if __name__ == "__main__":
    import sys

    if len(sys.argv) >= 2:
        print(list(map(lambda x: inriaMapTag(eval(x)), sys.argv[1:])))
    else:
        print("Need at least one argument to convert")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
""" Disjoint Set Data Structure for Kruskals Algorithm

    Author: Karthik Madathil <kmadathil@gmail.com>
"""
from copy import copy


class DisjointSet(object):
    ''' Disjoint Set Class with Indexing'''
    def __init__(self):
        self._data = {}

    def find(self, x):
        """
        Returns the representative member of the set of connected components to which x belongs, may be x itself.

        """
        if x not in self._data:
            self._data[x] = x
        if x != self._data[x]:
            self._data[x] = self.find(self._data[x])
        return self._data[x]

    def union(self, x, y):
        """
        Attaches the roots of x and y trees together if they are not the same already.
        :param x: first element
        :param y: second element
        :return: None
        """
        if x == y:
            return
        parent_x, parent_y = self.find(x), self.find(y)
        if parent_x != parent_y:
            self._data[parent_x] = parent_y

    def connected(self, x, y) -> bool:
        """
        :param x: first element
        :param y: second element
        :return: True if x and y belong to the same tree (i.e. they have the same root), False otherwise.
        """
        return self.find(x) == self.find(y)

    def copy(self):
        ''' Return a one-level deep copy '''
        t = DisjointSet()
        # for s in self:
        #     t.addset(copy(s))
        t._data = copy(self._data)
        return t



'''
Handles user-input and converts it into the format that internal modules expect
and vice-versa

@author:
'''

from __future__ import print_function, unicode_literals
import re
import logging

logger = logging.getLogger(__name__)

deletions = re.compile("[\u200b\u200c\u200d,'-;/().?!\"0123456789{}#\r]")


def normalize(s):
    """ Converts user-input into format expected by internal modules.
    Input s is expected to be an SLP1 encoded string
    """
    # Some bad visargas
    s = s.replace(':', 'H')
    # Remove Unicode Zero-Width characters, punctuation and numeric characters
    s = deletions.sub("", s)
    # Replace line-breaks with spaces
    s = s.replace('\n', ' ')
    if s[-1] == 'M':
        logger.debug("Detected anusvAra at end of string. Replacing with m")
        s = s[:-1] + 'm'
    if s[-1] == 'o':
        logger.debug("Detected o at end of string. Replacing with aH")
        s = s[:-1] + 'aH'
    return s


def replace_ending_visarga_s(s):
    """ Replace the final visarga of a string with s """
    if (s[-1] == 'H'):
        logger.debug("Detected H at end of string. Replacing with s")
        s = s[:-1] + 's'
    return s


def replace_ending_visarga_r(s):
    """ Replace the final visarga of a string with r """
    if (s[-1] == 'H'):
        logger.debug("Detected H at end of string. Replacing with s")
        s = s[:-1] + 'r'
    return s


def denormalize(s):
    """ Converts internal representation into user-friendly output
    Input s is expected to be an SLP1 encoded string
    """
    logger.debug("Denormalizing %s", s)
    s = s.replace("s ", "H ")
    s = s.replace("r ", "H ")
    s = re.sub("[sr]$", "H", s)
    logger.debug("%s", s)
    return s



'''
Created on Aug 4, 2017

@author: alvarna
'''

from sanskrit_parser.util.inriaxmlwrapper import InriaXMLWrapper
from sanskrit_parser.util.sanskrit_data_wrapper import SanskritDataWrapper
from sanskrit_parser.util.lexical_lookup import LexicalLookup
import logging


def _merge_tags(tags):
    ''' Merge tags from multiple sources

        Inputs
           tags: List of elements of form (baseword, tagset)
        Outputs
           list of elements of form (baseword, tagset), with
           tagsets properly merged
    '''
    tdict = {}
    # Convert to a dict of sets for proper set union
    for t in tags:
        base = t[0]
        if base not in tdict:
            tdict[base] = {frozenset(t[1])}
        else:
            ttags = frozenset(t[1])
            rs = set()
            add = True
            for tbs in tdict[base]:
                if ttags == tbs:  # Already in there
                    break
                elif ttags.issuperset(tbs):  # Superset, add and remove subset
                    rs.add(tbs)
                elif ttags.issubset(tbs):  # Subset, don't add
                    add = False
                    break
            if rs:
                tdict[base].difference_update(rs)
            if add:
                tdict[base].add(ttags)
    tlist = []
    # Convert back to list of tuples
    for base in tdict:
        tlist.extend([(base, set(s)) for s in tdict[base]])
    return tlist


class CombinedWrapper(LexicalLookup):

    def __init__(self, logger=None):
        self.inria = LexicalLookupFactory.create("inria")
        self.sanskrit_data = LexicalLookupFactory.create("sanskrit_data")
        self.logger = logger or logging.getLogger(__name__)

    def valid(self, word):
        return self.inria.valid(word) or self.sanskrit_data.valid(word)

    def get_tags(self, word, tmap=True):
        tags = self.inria.get_tags(word, tmap) or []
        sanskrit_data_tags = self.sanskrit_data.get_tags(word, tmap)
        if sanskrit_data_tags is not None:
            tags.extend(sanskrit_data_tags)
        tags = _merge_tags(tags)
        if tags == []:
            return None
        else:
            return tags


class LexicalLookupFactory(object):

    @staticmethod
    def create(name):
        if name == "inria":
            return InriaXMLWrapper()
        if name == "sanskrit_data":
            return SanskritDataWrapper()
        if name == "combined":
            return CombinedWrapper()
        raise Exception("invalid type", name)



#!/usr/bin/env python
"""
Wrapper around  kRShNamAchArya dhAtupATha to extract simple dhAtu attributes
@author: Karthikeyan Madathil (@kmadathil)
"""
from __future__ import print_function

import logging
import os

import requests
from tinydb import TinyDB, Query

from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SCHEMES


class DhatuWrapper(object):
    """
    Class to interface with the kRShNamAchArya dhAtupATha
    https://github.com/sanskrit-coders/stardict-sanskrit/tree/master/sa-vyAkaraNa/dhAtu-pATha-kRShNAchArya
    """
    # Moved to our own repo
    git_url = 'https://raw.githubusercontent.com/kmadathil/sanskrit_parser/master/data/dhAtu-pATha-kRShNAchArya.json'
    base_dir = os.path.expanduser("~/.sanskrit_parser/data")
    db_file = "dhAtu-pATha-kRShNAchArya.json"
    q = Query()

    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self._get_file()
        self.db = TinyDB(os.path.join(self.base_dir, self.db_file))
        # Check if db is empty
        if len(self.db.all()) == 0:
            self._generate_db()

    def _get_file(self):
        """ Download file if not present in cache """
        if not os.path.exists(self.base_dir):
            self.logger.debug("Data cache not found. Creating.")
            os.makedirs(self.base_dir)
        if not os.path.exists(os.path.join(self.base_dir, self.db_file)):
            self.logger.debug("%s not found. Downloading it", self.db_file)
            r = requests.get(self.git_url, stream=True)
            assert r.status_code != 404, \
                "Could not download file {}".format(self.git_url)
            with open(os.path.join(self.base_dir, self.db_file), "wb") as fd:
                for chunk in r.iter_content(chunk_size=128):
                    fd.write(chunk)

    def _get_dhatus(self, d):
        """ Get all tags for a dhatu d """
        if d is None:
            return None
        else:
            return self.db.search(self.q.DAtuH == d)

    def is_sakarmaka(self, d):
        """ Is d sakarmaka? """
        # Tags
        tl = self._get_dhatus(d)
        if len(tl) != 0:
            supported_karmakas = {'sakarmakaH', 'dvikarmakaH'}
            return any([t['karmakatvaM'] in supported_karmakas for t in tl])
        else:
            self.logger.debug("Couldn't find dhatu {} in database".format(d))
            return False


if __name__ == "__main__":
    from argparse import ArgumentParser

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = ArgumentParser(description='Dhatu Wrapper')
        # String to encode
        parser.add_argument('dhatu', nargs="?", type=str, default="kf")
        # Input Encoding (autodetect by default)
        parser.add_argument('--input-encoding', type=str, default=None)
        parser.add_argument('--tags', type=str,
                            choices=["all", u'DAtuH', u'mUlaDAtuH',
                                     u'DAtvarTaH', u'gaRaH', u'karmakatvaM',
                                     u'iwtvaM', u'padam-upagrahaH', u'rUpam'],
                            default=u'karmakatvaM')
        parser.add_argument('--debug', action='store_true')
        return parser.parse_args()

    def main():
        args = getArgs()
        print("Input Dhatu:", args.dhatu)
        if args.debug:
            logging.basicConfig(filename='DhatuWrapper.log', filemode='w', level=logging.DEBUG)
        else:
            logging.basicConfig(filename='DhatuWrapper.log', filemode='w', level=logging.INFO)
        logger = logging.getLogger(__name__)
        if args.input_encoding is None:
            ie = None
        else:
            ie = SCHEMES[args.input_encoding]
        i = SanskritImmutableString(args.dhatu, encoding=ie)
        it = i.canonical()
        print("Input String in SLP1:", it)
        logger.info("Input String in SLP1: {}".format(it))
        w = DhatuWrapper(logger=logger)
        if args.tags == "all":
            res = w._get_dhatus(it)
        else:
            res = map(lambda x: x[args.tags], w._get_dhatus(it))
        print(res)
        print("Is {} sakarmaka?: {}".format(it, w.is_sakarmaka(it)))
        logger.info("Reported {}".format(res))

    main()



# -*- coding: utf-8 -*-
"""
Intro
=====
Sandhi splitter for Samskrit.
Builds up a database of sandhi rules and utilizes them for both
performing sandhi and splitting words.

Will generate splits that may not all be valid words. That is left to the
calling module to validate. See for example SanskritLexicalAnalyzer

Example usage:
    from sandhi import Sandhi
    sandhi = Sandhi()
    joins = sandhi.join('tasmin', 'iti')
    splits = sandhi.split_at('tasminniti', 5)

Draws inspiration from https://github.com/sanskrit/sanskrit

@author: Avinash Varna (github: @avinashvarna)

Usage
=====
The ``Sandhi`` class can be used to join/split words:

.. code:: python

    >>> from sanskrit_parser.parser.sandhi import Sandhi
    >>> sandhi = Sandhi()
    >>> word1 = SanskritImmutableString('te')
    >>> word2 = SanskritImmutableString('eva')
    >>> joins = sandhi.join(word1, word2)
    >>> for join in joins:
    ...    print(join)
    ...
    teeva
    taeva
    ta eva
    tayeva

To split at a specific position, use the ``Sandhi.split_at()`` method:

.. code:: python

    >>> w = SanskritImmutableString('taeva')
    >>> splits = sandhi.split_at(w, 1)
    >>> for split in splits:
    ...    print(split)
    ...
    (u'tar', u'eva')
    (u'tas', u'eva')
    (u'taH', u'eva')
    (u'ta', u'eva')

To split at all possible locations, use the ``Sandhi.split_all()``
method:

.. code:: python

    >>> splits_all = sandhi.split_all(w)
    >>> for split in splits_all:
    ...    print(split)
    ...
    (u't', u'aeva')
    (u'tar', u'eva')
    (u'taev', u'a')
    (u'to', u'eva')
    (u'ta', u'eva')
    (u'te', u'eva')
    (u'taH', u'eva')
    (u'tae', u'va')
    (u'taeva', u'')
    (u'tas', u'eva')

**Note**: As mentioned previously, both over-generation and
under-generation are possible with the ``Sandhi`` class.


Command line usage
==================

::

    $ python -m sanskrit_parser.parser.sandhi --join te eva
    Joining te eva
    set([u'teeva', u'taeva', u'ta eva', u'tayeva'])

    $ python -m sanskrit_parser.parser.sandhi --split taeva 1
    Splitting taeva at 1
    set([(u'tar', u'eva'), (u'tas', u'eva'), (u'taH', u'eva'), (u'ta', u'eva')])

    $ python -m sanskrit_parser.parser.sandhi --split taeva --all
    All possible splits for taeva
    set([(u't', u'aeva'), (u'tar', u'eva'), (u'taev', u'a'), (u'to', u'eva'),
    (u'ta', u'eva'), (u'te', u'eva'), (u'taH', u'eva'), (u'tae', u'va'),
    (u'taeva', u''), (u'tas', u'eva')])



"""

from __future__ import print_function
from collections import defaultdict
import itertools
import codecs
import os
import re
import inspect
import logging
import datetime
import six
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SanskritNormalizedString, SLP1, SCHEMES, outputctx
from sanskrit_parser.base.maheshvara_sutra import MaheshvaraSutras


class Sandhi(object):
    """
    Class to hold all the sandhi rules and methods for joining and splitting.
    Uses SLP1 encoding for all internal operations.
    """

    def __init__(self, rules_dir=None, use_default_rules=True, logger=None):
        """
        Sandhi class constructor

        :param rules_dir: directory to read rules from
        :param use_default_rules: reads pre-built-rules from sandhi_rules dir under module directory
        :param logger: instance of python logger to use
        """
        self.forward = defaultdict(set)
        self.backward = defaultdict(set)
        self.lc_len_max = 0
        self.rc_len_max = 0
        self.after_len_max = 0
        self.logger = logger or logging.getLogger(__name__)
        if rules_dir:
            self.add_rules_from_dir(rules_dir)
        if use_default_rules:
            base_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
            self.add_rules_from_dir(os.path.join(base_dir, "sandhi_rules"))

    def add_rule(self, before, after, annotation=None):
        """
        Adds a single sandhi rule to the rule database

        :param before: tuple of length two of the varNas involved in the sandhi
        :param after: result of the sandhi
        """
        self.forward[before].add((after, annotation))
        self.backward[after].add((before, annotation))
        if len(before[0]) > self.lc_len_max:
            self.logger.debug("Setting lc_len_max to %d", len(before[0]))
        if len(before[1]) > self.rc_len_max:
            self.logger.debug("Setting rc_len_max to %d", len(before[1]))
        if len(after) > self.after_len_max:
            self.logger.debug("Setting after_len_max to %d", len(after))
        self.lc_len_max = max(self.lc_len_max, len(before[0]))
        self.rc_len_max = max(self.rc_len_max, len(before[1]))
        self.after_len_max = max(self.after_len_max, len(after))

    def join(self, first_in, second_in):
        """
        Performs sandhi.
        **Warning**: May generate forms that are not lexically valid.

        :param first_in: SanskritImmutableString first word of the sandhi
        :param second_in: SanskritImmutableString word of the sandhi
        :return: list of strings of possible sandhi forms, or None if no sandhi can be performed
        """
        first = first_in.canonical()
        second = second_in.canonical()
        self.logger.debug("Join: {}, {}".format(first, second))
        if first is None or len(first) == 0:
            return second
        if second is None:
            return first
        left_chars = [first[i:] for i in range(max(0, len(first)-self.lc_len_max), len(first))]
        left_chars.append("^"+first)
        right_chars = [second[0:i] for i in range(min(self.rc_len_max, len(second))+1)]
        self.logger.debug("left_chars = %s, right_chars %s", left_chars, right_chars)
        joins = set()
        for key in itertools.product(left_chars, right_chars):
            afters = self.forward.get(key)
            if afters:
                for after, annotation in afters:
                    self.logger.debug("Found sandhi %s = %s (%s)", key, after, annotation)
                    joins.add(first[:-len(key[0])] + after + second[len(key[1]):])
        if len(joins) == 0:
            self.logger.debug("No joins found")
            return None
        else:
            return joins

    def split_at(self, word_in, idx):
        """
        Split sandhi at the given index of word.
        **Warning**: Will generate splits that are not lexically valid.

        :param word_in: SanskritImmutableString word to split
        :param idx: position within word at which to try the split
        :return: set of tuple of strings of possible split forms, or None if no split can be performed
        """
        word = word_in.canonical()
        self.logger.debug("Split: %s, %d", word, idx)
        splits = set()
        # Figure out how may chars we can extract for the afters
        stop = min(idx+self.after_len_max, len(word))
        afters = [word[idx:i] for i in range(idx+1, stop+1)]
        for after in afters:
            self.logger.debug("Trying after %s", after)
            befores = self.backward[after]
            if befores:
                for before, annotation in befores:
                    self.logger.debug("Found split %s -> %s (%s)", after, before, annotation)
                    # Do we have a beginning-of-line match rule
                    if before[0][0] == "^":
                        if idx != 0:
                            # Can't allow matches at any other position
                            continue
                        else:
                            # drop the ^ in the result
                            before = (before[0][1:], before[1])
                    left = word[:idx] + before[0]
                    right = before[1] + word[idx+len(after):]
                    splits.add((left, right))

        if len(splits) == 0:
            self.logger.debug("No split found")
            return None
        else:
            return splits

    def split_all(self, word_in, start=None, stop=None):
        """
        Split word at all possible locations and return splits.
        **Warning**: Will generate splits that are not lexically valid.

        :param word_in: SanskritImmutableString word to split
        :return: set of tuple of strings of possible split forms, or None if no split can be performed
        """
        splits = set()
        word = word_in.canonical()
        start = start or 0
        stop = stop or len(word)
        for idx in six.moves.range(start, stop):
            split = self.split_at(word_in, idx)
            if split:
                splits |= split
        if len(splits) == 0:
            self.logger.debug("No split found")
            return None
        else:
            return splits

    def expand_rule(self, rule):
        """
        Expands a given sandhi rule from the rules file to generate all possible combinations

        :param rule: Rule to expand
        :return: A generator of all possible expanded rules
        """
        self.logger.debug("Expanding rule %s", rule)

        ms = MaheshvaraSutras()

        b, afters = map(six.text_type.strip, rule.split("="))
        before = list(map(six.text_type.strip, b.split("+", 1)))
        left_classes = re.split(r'\[(.*?)\]', before[0])
        self.logger.debug("Left classes = %s", left_classes)

        # Split after forms into individual forms
        afters = map(six.text_type.strip, afters.split("/"))

        before_left = []
        for c in left_classes:
            if c != '':
                if c.startswith("*"):
                    # This is a mAheswara sUtra pratyAhAra
                    splits = list(map(six.text_type.strip, c.split('-')))
                    varnas = set(ms.getPratyahara(
                        SanskritImmutableString(splits[0][1:], encoding=SLP1),
                        longp=False, remove_a=True, dirghas=True).canonical())
                    if len(splits) == 2:
                        varnas -= set(splits[1])
                    self.logger.debug("Found pratyAhAra %s = %s", c, varnas)
                    before_left.append(varnas)
                else:
                    before_left.append(map(six.text_type.strip, c.split(",")))
        self.logger.debug("before_left iterator = %s", before_left)

        right_classes = re.split(r'\[(.*?)\]', before[1])
        # Could have used list comprehension, but this is easier to read
        self.logger.debug("right_classes = %s", right_classes)
        if right_classes:
            before_right = []
            for c in right_classes:
                if c != '':
                    if c.startswith("*"):
                        # This is a mAheswara sUtra pratyAhAra
                        splits = list(map(six.text_type.strip,
                                          re.split('([+-])', c)))
                        varnas = set(
                            ms.getPratyahara(
                                SanskritImmutableString(splits[0][1:], encoding=SLP1),
                                longp=False, remove_a=True, dirghas=True)
                            .canonical()
                        )
                        if len(splits) == 3:
                            if splits[1] == '-':
                                varnas -= set(splits[2])
                            elif splits[1] == '+':
                                varnas |= set(splits[2])
                        self.logger.debug("Found pratyAhAra %s (%s) = %s", c, splits[0][1:], varnas)
                        before_right.append(varnas)
                    else:
                        before_right.append(map(six.text_type.strip, c.split(",")))
        else:
            before_right = [before[1].strip()]
        self.logger.debug("before_right iterator = %s", before_right)

        for after, before_l, before_r in itertools.product(afters,
                                                           itertools.product(*before_left),
                                                           itertools.product(*before_right)):
            left = ''.join(before_l)
            right = ''.join(before_r)
            list_before_r = list(before_r)
            left_right = (left, right)
            a = after.format(*(list(before_l) + list_before_r))
            # The below is just too much logging - should be silenced in production:
            # self.logger.debug("Final rule = %s -> %s", left_right, a)
            yield (left_right, a)

    def add_rules_from_file(self, path):
        """
        Add sandhi rules from file.
        Each line of the input file should contain one rule. E.g. अ + अ = आ
        Lines starting with a # are treated as comments and skipped.
        Empty lines are ignored as well.

        :param path: file to read rules from

        See also add_rules_from_dir
        """
        filename = os.path.basename(path)
        with codecs.open(path, "rb", 'utf-8') as f:
            for linenum, line in enumerate(f):
                line = line.strip()
                if line.startswith('#') or line == '':
                    continue
                self.logger.debug("Processing rule %s", line)
                rule = SanskritImmutableString(line).canonical()
                for r in self.expand_rule(rule):
                    self.add_rule(*r, annotation="%s:%d" % (filename, linenum+1))

    def add_rules_from_dir(self, directory):
        """
        Add sandhi rules from an entire directory.
        Reads all .txt files from the given directory and adds rules from them


        :param directory: path to directory with rules files

        See also add_rules_from_file
        """
        self.logger.debug("Adding rules from directory %s", directory)
        for filename in os.listdir(directory):
            if filename.endswith(".txt"):
                self.logger.debug("Processing rules from file %s", filename)
                self.add_rules_from_file(os.path.join(directory, filename))
            else:
                continue


if __name__ == "__main__":
    from argparse import ArgumentParser

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = ArgumentParser(description='Sandhi Utility')
        # Input Encoding (autodetect by default)
        parser.add_argument('--input-encoding', type=str, default=None)
        parser.add_argument('--loglevel', type=str, help="logging level. Can be any level supported by logging module")
        parser.add_argument('--split', action='store_true', help="Split the given word using sandhi rules")
        parser.add_argument('--join', action='store_true', help="Join the given words using sandhi rules")
        parser.add_argument('--all', action='store_true', help="Return splits at all possible locations")
        parser.add_argument('--strict-io', action='store_true',
                            help="Do not modify the input/output string to match conventions", default=False)

        # String to encode
        parser.add_argument('word', nargs='?', type=str,
                            default="tasminniti",
                            help="First word of sandhi if join, or word to split")
        parser.add_argument('word_or_pos', nargs="?", type=str,
                            default="eva",
                            help="Second word of sandhi if join, or position to split")

        return parser.parse_args()

    def main():
        args = getArgs()
        if args.input_encoding is None:
            ie = None
        else:
            ie = SCHEMES[args.input_encoding]

        # Setup logging
        if args.loglevel:
            numeric_level = getattr(logging, args.loglevel.upper(), None)
            if not isinstance(numeric_level, int):
                raise ValueError('Invalid log level: %s' % args.loglevel)
            logging.basicConfig(filename="sandhi.log", filemode="wb", level=numeric_level)

        logging.info("---------------------------------------------------")
        logging.info("Started processing at %s", datetime.datetime.now())

        sandhi = Sandhi()
        # if neither split nor join is chosen, just demo both
        if not args.split and not args.join:
            print("Neither split nor join option chosen. Here's a demo of joining")
            args.join = True
        with outputctx(args.strict_io):
            if args.split:
                word_in = SanskritNormalizedString(args.word, encoding=ie, strict_io=args.strict_io)
                if args.all:
                    print("All possible splits for {}".format(args.word))
                    splits = sandhi.split_all(word_in)
                else:
                    pos = int(args.word_or_pos)
                    print("Splitting {0} at {1}".format(args.word, pos))
                    splits = sandhi.split_at(word_in, pos)
                print(splits)
            if args.join:
                print("Joining {0} {1}".format(args.word, args.word_or_pos))
                first_in = SanskritNormalizedString(args.word, encoding=ie, strict_io=args.strict_io)
                second_in = SanskritNormalizedString(args.word_or_pos, encoding=ie, strict_io=args.strict_io)
                joins = sandhi.join(first_in, second_in)
                print(joins)

        logging.info("Finished processing at %s", datetime.datetime.now())
        logging.info("---------------------------------------------------")
        logging.shutdown()

    main()



# -*- coding: utf-8 -*-
"""

Usage
======

The ``Parser`` class can be used to generate vakya parses thus:

.. code-block:: python

        from itertools import islice
        from sanskrit_parser import Parser
        string = "astyuttarasyAMdiSi"
        input_encoding = "SLP1"
        output_encoding = "SLP1"
        parser = Parser(input_encoding=input_encoding,
                        output_encoding=output_encoding,
                        replace_ending_visarga='s')
        parse_result = parser.parse(string)
        print('Splits:')
        for split in parse_result.splits(max_splits=10):
            print(f'Lexical Split: {split}')
            for i, parse in enumerate(islice(split.parses(), 3)):
                print(f'Parse {i}')
                print(f'{parse}')
        ...
Lexical Split: ['asti', 'uttarasyAm', 'diSi']
Parse 0
asti => (asti, ['samAsapUrvapadanAmapadam', 'strIliNgam']) : samasta of uttarasyAm
uttarasyAm => (uttara#1, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam'])
diSi => (diS, ['saptamIviBaktiH', 'ekavacanam', 'strIliNgam']) : viSezaRa of uttarasyAm
Parse 1
asti => (asti, ['samAsapUrvapadanAmapadam', 'strIliNgam']) : samasta of uttarasyAm
uttarasyAm => (uttara#2, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam']) : viSezaRa of diSi
diSi => (diS#2, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam'])
Parse 2
asti => (as#1, ['kartari', 'praTamapuruzaH', 'law', 'parasmEpadam', 'ekavacanam', 'prATamikaH'])
uttarasyAm => (uttara#2, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam']) : viSezaRa of diSi
diSi => (diS, ['saptamIviBaktiH', 'ekavacanam', 'strIliNgam']) : aDikaraRam of asti


@author: avinashvarna
"""

import time
import json
import abc
from dataclasses import dataclass
from typing import Sequence
from sanskrit_parser.base.sanskrit_base import SCHEMES, SanskritObject, SLP1
from sanskrit_parser.base.sanskrit_base import SanskritNormalizedString, SanskritString
from sanskrit_parser.parser.sandhi_analyzer import LexicalSandhiAnalyzer
from sanskrit_parser.parser.datastructures import VakyaGraph, VakyaGraphNode
from sanskrit_parser.parser.datastructures import SandhiGraph
import logging

logger = logging.getLogger(__name__)


class Serializable(abc.ABC):
    """ Base class to indicate an object is serializable into JSON """

    @abc.abstractmethod
    def serializable(self):
        ''' Return an object that can be serialized by json.JSONEncoder '''
        pass


class JSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, Serializable):
            return o.serializable()
        return super().default(o)


@dataclass
class Parser():
    strict_io: bool = False
    input_encoding: str = None
    output_encoding: str = 'SLP1'
    lexical_lookup: str = "combined"
    score: bool = True
    split_above: int = 5
    replace_ending_visarga: str = None
    fast_merge = True

    def parse(self,
              input_string: str,
              ):
        if self.input_encoding is not None:
            self.input_encoding = SCHEMES[self.input_encoding]
        if self.output_encoding is not None:
            self.output_encoding = SCHEMES[self.output_encoding]
        s = SanskritNormalizedString(input_string,
                                     encoding=self.input_encoding,
                                     strict_io=self.strict_io,
                                     replace_ending_visarga=self.replace_ending_visarga)
        logger.info(f"Input String in SLP1: {s.canonical()}")
        sandhi_analyzer = LexicalSandhiAnalyzer(self.lexical_lookup)
        logger.debug("Start Split")
        graph = sandhi_analyzer.getSandhiSplits(s, tag=True)
        logger.debug("End DAG generation")
        return ParseResult(self, input_string, graph)


@dataclass
class ParseResult(Serializable):
    parser: Parser
    input_string: str
    sandhi_graph: SandhiGraph

    def __str__(self):
        return f"ParseResult('{self.input_string}')"

    def splits(self, max_splits: int = 10):
        split_results = self.sandhi_graph.find_all_paths(max_paths=max_splits,
                                                         sort=True,
                                                         score=self.parser.score)
        logger.debug("End pathfinding")
        logger.debug("Splits:")
        for split in split_results:
            logger.debug(f"Lexical Split: {split}")
            yield Split(self.parser, self.input_string, split)

    def serializable(self):
        return list(self.splits(10))

    def write_dot(self, basepath):
        self.sandhi_graph.write_dot(basepath)


@dataclass
class Split(Serializable):
    parser: Parser
    input_string: str
    split: Sequence[SanskritObject]
    vgraph: VakyaGraph = None

    def __repr__(self):
        return f'Split({self.input_string}) = {self.split}'

    def __str__(self):
        strict_io = self.parser.strict_io
        encoding = self.parser.output_encoding
        out = [t.transcoded(encoding, strict_io) for t in self.split]
        return str(out)

    def parses(self):
        self.vgraph = VakyaGraph(self.split,
                                 fast_merge=self.parser.fast_merge,
                                 max_parse_dc=self.parser.split_above)
        for (ix, parse_graph) in enumerate(self.vgraph.parses):
            logger.debug(f"Parse {ix}")
            yield Parse(self, parse_graph)

    def write_dot(self, basepath):
        self.vgraph.write_dot(basepath)

    def serializable(self):
        strict_io = self.parser.strict_io
        encoding = self.parser.output_encoding
        out = [t.transcoded(encoding, strict_io) for t in self.split]
        return {'split': out,
                'parses': list(self.parses())
                }


@dataclass
class ParseTag(Serializable):
    root: str
    tags: Sequence[str]

    def serializable(self):
        return {'root': self.root, 'tags': self.tags}

    def __str__(self):
        return f'({self.root}, {self.tags})'


class ParseNode(Serializable):
    def __init__(self, node: VakyaGraphNode,
                 strict_io: bool,
                 encoding: str):
        self.pada = node.pada.transcoded(encoding, strict_io)
        tag = node.getMorphologicalTags()
        self.parse_tag = ParseTag(tag[0].transcoded(encoding, strict_io),
                                  [t.transcoded(encoding, strict_io) for t in tag[1]]
                                  )

    def __str__(self):
        return f'{self.pada} => {self.parse_tag}'

    def serializable(self):
        d = {'pada': self.pada}
        d.update(self.parse_tag.serializable())
        return d


@dataclass
class ParseEdge(Serializable):
    predecessor: str
    node: ParseNode
    label: str

    def __str__(self):
        return f'{self.node} : {self.label} of {self.predecessor}'

    def serializable(self):
        return {'node': self.node,
                'predecessor': self.predecessor,
                'sambandha': self.label
                }


class Parse(Serializable):

    def __init__(self, split: Split, parse_graph):
        strict_io = split.parser.strict_io
        encoding = split.parser.output_encoding
        graph = []
        for n in sorted(list(parse_graph), key=lambda x: x.index):
            node = ParseNode(n, strict_io, encoding)
            preds = list(parse_graph.predecessors(n))
            if preds:
                pred = preds[0]  # Only one
                lbl = parse_graph.edges[pred, n]['label']
                edge = ParseEdge(pred.pada.transcoded(encoding, strict_io),
                                 node,
                                 SanskritString(lbl, encoding=SLP1).transcoded(encoding, strict_io)
                                 )
                graph.append(edge)
            else:
                graph.append(node)
        self.graph = graph

    def __str__(self):
        return '\n'.join([str(t) for t in self.graph])

    def serializable(self):
        return {'graph': self.graph}


if __name__ == "__main__":
    start_time = time.time()

    def api_example(string, output_encoding):
        from itertools import islice
        parser = Parser(output_encoding=output_encoding,
                        replace_ending_visarga='s')
        parse_result = parser.parse(string)
        print('Splits:')
        for split in parse_result.splits(max_splits=10):
            print(f'Lexical Split: {split}')
            for i, parse in enumerate(islice(split.parses(), 2)):
                print(f'Parse {i}')
                print(f'{parse}')
            break
        print(json.dumps(split,
                         ensure_ascii=False,
                         indent=2,
                         cls=JSONEncoder)
              )

    def main():
        examples = [('devadattogrAmaMgacCati', 'SLP1'),
                    ('astyuttarasyAMdishidevatAtmA', 'Devanagari')
                    ]
        for string, encoding in examples:
            api_example(string, encoding)

    main()
    print(f'Took {time.time() - start_time} s')



#!/usr/bin/python3 -u

# This web app may be run in two modes. See bottom of the file.

import logging
import os.path
import sys

from sanskrit_parser.rest_api.flask_helper import app
from sanskrit_parser.rest_api import api_v1

logging.basicConfig(
     level=logging.DEBUG,
     filename="SanskritParserApi.log",
     filemode='w',
     format="%(levelname)s: %(asctime)s {%(filename)s:%(lineno)d}: %(message)s "
)

console = logging.StreamHandler()
console.setLevel(logging.INFO)
# set a format which is simpler for console use
formatter = logging.Formatter('%(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger(' ').addHandler(console)


# Add parent directory to PYTHONPATH, so that sanskrit_parser module can be found.
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
logging.debug(sys.path)

params = {
    'port': 9000,
}


def setup_app():
    app.register_blueprint(api_v1.api_blueprint, url_prefix="/sanskrit_parser")


def main(argv):
    setup_app()
    app.run(
      host="0.0.0.0",
      debug=False,
      port=params["port"],
      use_reloader=False
    )


if __name__ == "__main__":
    logging.info("Running in stand-alone mode.")
    main(sys.argv[1:])
else:
    logging.info("Likely running as a WSGI app.")
    setup_app()



import logging
import os
from base64 import b64encode

import flask
import jsonpickle
# from flask import url_for
from flask_cors import CORS


""" The flask app we serve in run.py.
"""
app = flask.Flask(
  # We pass the root module name - sets root directory.
  import_name="sanskrit_parser.rest_api.run")

# Let Javascsipt hosted elsewhere access our API.
CORS(app=app,
     # injects the `Access-Control-Allow-Credentials` header in responses. This allows cookies and credentials to be submitted across domains.
     supports_credentials=True)
logging.info(str(app))

app.config.update(
  DEBUG=True,
  # Used to encrypt session cookies.
  SECRET_KEY=b64encode(os.urandom(24)).decode('utf-8'),
)


@app.route('/')
def index():
    flask.session['logstatus'] = 1
    return flask.redirect('sanskrit_parser/docs')


# Cant use flask-sitemap - won't list flask restplus routes.
@app.route("/sitemap")
def site_map():
    output = []
    for rule in app.url_map.iter_rules():
        options = {}
    for arg in rule.arguments:
        options[arg] = "[{0}]".format(arg)

    methods = ','.join(rule.methods)
    url = str(rule)
    import urllib.request

    line = urllib.request.unquote("{:50s} {:20s} {}".format(rule.endpoint, methods, url))
    output.append(line)

    logging.info(str(output))
    response = app.response_class(
      response=jsonpickle.dumps(output),
      status=200,
      mimetype='application/json'
    )
    return response



"""Intro
======

Sanskrit parser is a python library to help parse Sanskrit input

It provides three main levels of output, in order of increasing complexity:
  1. *tags* - Morphological analysis of a word
  2. *sandhi* - Sandhi Split of a phrase
  3. *vakya* - Morphological Analysis of a sentence (after Sandhi split)

This project is still under development. Please report any issues `here
<https://github.com/kmadathil/sanskrit_parser/issues>`_.

Command line usage
==================

The ``sanskrit_parser`` script can be used to view parses, sandhi splits,
or morphological tags as below.

If the ``--dot`` option is provided, a graph is output in ``.dot`` fomat with
all the possible morphologies as nodes, and possible relations as
edges. The valid parses extracted from this graph are also written out
as ``_parse.dot`` files


::

    $ sanskrit_parser vakya astyuttarasyAMdiSi --input SLP1 --dot vakya.dot
    ...
    Lexical Split: [asti, uttarasyAm, diSi]
    ...
    Parse 0
    asti=>['as#1', {prATamikaH, praTamapuruzaH, kartari, ekavacanam, law}]
    diSi=>['diS#2', {ekavacanam, strIliNgam, saptamIviBaktiH}]
    uttarasyAm=>['uttara#1', {ekavacanam, strIliNgam, saptamIviBaktiH}]
    ...

    $ dot -Tpng vakya.dot -o vakya.png
    $ eog vakya.png
    $ dot -Tpng vakya_parse0.dot -o vakya.png
    $ eog vakya_parse0.png


::

    $ sanskrit_parser sandhi astyuttarasyAMdishidevatAtmA
    ...
    Splits:
    [asti, uttarasyAm, diSi, devatA, AtmA]
    [asti, uttara, syAm, diSi, devatA, AtmA]
    [astI, uttarasyAm, diSi, devatA, AtmA]
    [asti, uttarasyAm, diSi, devatAt, mA]
    [asti, uttarasyAm, diSi, devata, AtmA]
    [asti, uttarasyAm, diSi, devat, AtmA]
    [asti, uttarasyAm, diSi, devatAtmA]
    [asti, uttarasyAm, diSi, devatA, at, mA]
    [asti, uttara, syAm, diSi, devat, AtmA]
    [asti, uttarasyAm, di, Si, devatA, AtmA]
    -----------
    ..

::

    $ scripts/sanskrit_parser tags hares
    ...
    Morphological tags:
    ('hf#1', {ekavacanam, viDiliN, prATamikaH, kartari, maDyamapuruzaH})
    ('hari#1', {ekavacanam, puMlliNgam, paYcamIviBaktiH})
    ('hari#1', {ekavacanam, paYcamIviBaktiH, strIliNgam})
    ('hari#1', {zazWIviBaktiH, ekavacanam, puMlliNgam})
    ('hari#1', {zazWIviBaktiH, ekavacanam, strIliNgam})
    ('hari#2', {ekavacanam, puMlliNgam, paYcamIviBaktiH})
    ('hari#2', {ekavacanam, paYcamIviBaktiH, strIliNgam})
    ('hari#2', {zazWIviBaktiH, ekavacanam, puMlliNgam})
    ('hari#2', {zazWIviBaktiH, ekavacanam, strIliNgam})
    ('hf', {ekavacanam, parasmEpadam, viDiliN, maDyamapuruzaH})
    ('hari', {ekavacanam, puMlliNgam, paYcamIviBaktiH})
    ('hari', {ekavacanam, strIliNgam, paYcamIviBaktiH})
    ('hari', {zazWIviBaktiH, ekavacanam, puMlliNgam})
    ('hari', {zazWIviBaktiH, ekavacanam, strIliNgam})

"""
from .api import Parser
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def enable_console_logger(level=logging.INFO,
                          fmt='%(levelname)-8s %(message)s'):
    '''Enable logger Console Logging for sanskrit_parser

       Params
           level: log level
           fmt  : log format
    '''
    logger = logging.getLogger(__name__)
    console = logging.StreamHandler()
    console.setLevel(level)
    # set a format which is simpler for console use
    formatter = logging.Formatter(fmt)
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logger.addHandler(console)


def enable_file_logger(log_file_name='SanskritParser.log',
                       level=logging.INFO,
                       fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s'):
    '''Enable logger File Logging for sanskrit_parser

       Params
           log_file_name(str): Log file Name
           level: log level
           fmt  : log format
    '''
    logger = logging.getLogger(__name__)
    # create file handler which logs even debug messages
    formatter = logging.Formatter(fmt)
    fh = logging.FileHandler(log_file_name)
    fh.setLevel(level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


__all__ = ['Parser']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
""" Sanskrit Parser Data Structures

    @author: Karthik Madathil (github: @kmadathil)
"""

from sanskrit_parser.base.sanskrit_base import SanskritObject, SanskritImmutableString, SLP1
import networkx as nx
from itertools import islice
import logging
import operator
from copy import copy
import six
import time
from collections import defaultdict
from os.path import dirname, basename, splitext, join
from sanskrit_parser.util import lexical_scorer
from sanskrit_parser.util.disjoint_set import DisjointSet
from sanskrit_parser.util.DhatuWrapper import DhatuWrapper

__all__ = ['SandhiGraph', 'VakyaGraph', 'VakyaParse', 'getSLP1Tagset']

dw = DhatuWrapper()


logger = logging.getLogger(__name__)


class SandhiGraph(object):
    """ DAG class to hold Sandhi Lexical Analysis Results

        Represents the results of lexical sandhi analysis as a DAG
        Nodes are SanskritObjects
    """
    start = "__start__"
    end = "__end__"

    def __init__(self):
        ''' DAG Class Init

        Params:
            elem (SanskritObject :optional:): Optional initial element
            end  (bool :optional:): Add end edge to initial element
        '''
        self.roots = []
        self.G = nx.DiGraph()
        self.scorer = lexical_scorer.Scorer()

    def __iter__(self):
        ''' Iterate over nodes '''
        return self.G.__iter__()

    def has_node(self, t):
        ''' Does a given node exist in the graph?

            Params:
               t (SanskritObject): Node
            Returns:
               boolean
        '''
        return t in self.G

    def append_to_node(self, t, nodes):
        """ Create edges from t to nodes

            Params:
                t (SanskritObject)      : Node to append to
                nodes (iterator(nodes)) : Nodes to append to t
        """
        # t is in our graph
        assert t in self.G
        for r in nodes:
            self.G.add_edge(t, r)

    def add_node(self, node):
        """ Extend dag with node inserted at root

            Params:
                Node (SanskritObject)      : Node to add
                root (Boolean)             : Make a root node
                end  (Boolean)             : Add an edge to end
        """
        assert node not in self.G
        self.G.add_node(node)

    def add_end_edge(self, node):
        ''' Add an edge from node to end '''
        assert node in self.G
        self.G.add_edge(node, self.end)

    def add_roots(self, roots):
        self.roots.extend(roots)

    def lock_start(self):
        ''' Make the graph ready for search by adding a start node

        Add a start node, add arcs to all current root nodes, and clear
        self.roots
        '''
        self.G.add_node(self.start)
        for r in self.roots:
            self.G.add_edge(self.start, r)
        self.roots = []

    def score_graph(self):
        edges = self.G.edges()
        edges_list = []
        edges_to_score = []
        for edge in edges:
            edges_list.append(edge)
            if edge[0] == self.start:
                edges_to_score.append([edge[1]])
            elif edge[1] == self.end:
                edges_to_score.append([edge[0]])
            else:
                edges_to_score.append(edge)
        scores = self.scorer.score_splits(edges_to_score)
        for edge, score in zip(edges_list, scores):
            # Score is log-likelihood, so higher is better.
            # For graph path-finding, smaller weight is better, so use negative
            edges[edge]['weight'] = -score
        for u, v, w in self.G.edges(data='weight'):
            logger.debug("u = %s, v = %s, w = %s", u, v, w)

    def find_all_paths(self, max_paths=10, sort=True, score=True):
        """ Find all paths through DAG to End

            Params:
               max_paths (int :default:=10): Number of paths to find
                          If this is > 1000, all paths will be found
               sort (bool)                 : If True (default), sort paths
                                             in ascending order of length
        """
        if self.roots:
            self.lock_start()
        if score:
            self.score_graph()
        # shortest_simple_paths is slow for >1000 paths
        if max_paths <= 1000:
            if score:
                paths = list(six.moves.map(lambda x: x[1:-1],
                                           islice(nx.shortest_simple_paths(
                                                        self.G, self.start, self.end, weight='weight'),
                                                  max_paths)))
                scores = self.scorer.score_splits(paths)
                path_scores = zip(paths, scores)
                sorted_path_scores = sorted(path_scores, key=operator.itemgetter(1), reverse=True)
                logger.debug("Sorted paths with scores:\n %s", sorted_path_scores)
                # Strip the scores from the returned result, to be consistent with no-scoring option
                sorted_paths, _ = zip(*sorted_path_scores)
                return list(sorted_paths)
            else:
                paths = list(six.moves.map(lambda x: x[1:-1],
                                           islice(nx.shortest_simple_paths(
                                                        self.G, self.start, self.end),
                                                  max_paths)))
                return paths
        else:  # Fall back to all_simple_paths
            ps = list(six.moves.map(lambda x: x[1:-1],
                                    nx.all_simple_paths(self.G, self.start, self.end)))
            # If we do not intend to display paths, no need to sort them
            if sort:
                ps.sort(key=lambda x: len(x))
            return ps

    def __str__(self):
        """ Print representation of DAG """
        return str(self.G)

    def draw(self, *args, **kwargs):
        _ncache = {}

        def _uniq(s):
            if s not in _ncache:
                _ncache[s] = 0
                return s
            else:
                _ncache[s] = _ncache[s] + 1
                return s + "_" + str(_ncache[s])

        nx.draw(self.G, *args, **kwargs,
                pos=nx.spring_layout(self.G),
                labels={x: _uniq(str(x)) for x in self})

    def write_dot(self, path):
        # H = nx.convert_node_labels_to_integers(G, label_attribute=’node_label’)
        # H_layout = nx.nx_pydot.pydot_layout(G, prog=’dot’)
        # G_layout = {H.nodes[n][‘node_label’]: p for n, p in H_layout.items()}
        nx.drawing.nx_pydot.write_dot(self.G, path)


lakaras = set(['law', 'liw', 'luw', 'lrw', 'low', 'laN', 'liN', 'luN', 'lfN',
               'viDiliN', 'ASIrliN'])
krtverbs = set(['ktvA', 'Satf', 'Sanac', 'tumun', 'kta', 'ktavatu', 'lyap'])
purvakala = set(['ktvA', 'lyap'])
samanakala = set(['Satf', 'Sanac'])
nishta = set(['kta', 'ktavatu'])
karmani = set(['karmaRi'])
samastas = set(['samAsapUrvapadanAmapadam'])
# Vibhaktis
_vibhaktis = ['praTamAviBaktiH', 'dvitIyAviBaktiH', 'tftIyAviBaktiH',
              'caturTIviBaktiH', 'paYcamIviBaktiH', 'zazWIviBaktiH',
              'saptamIviBaktiH', 'saMboDanaviBaktiH']
prathama = _vibhaktis[1-1]
dvitiya = _vibhaktis[2-1]
tritiya = _vibhaktis[3-1]
chaturthi = _vibhaktis[4-1]
pancami = _vibhaktis[5-1]
shashthi = _vibhaktis[6-1]
saptami = _vibhaktis[7-1]
sambodhana = _vibhaktis[8-1]
vibhaktis = set(_vibhaktis)
# Vacanas
vacanas = set(['ekavacanam', 'dvivacanam', 'bahuvacanam'])
# Puruzas
puruzas = 'praTamapuruzaH', 'maDyamapuruzaH', 'uttamapuruzaH'
karakas = set(['kartA', 'karma', 'karaRam', 'apAdAnam', 'sampradAnam', 'aDikaraRam'])
predicative_verbs = set(['as', 'BU', 'vft'])
_lingas = ['puMlliNgam', 'napuMsakaliNgam', 'strIliNgam', 'triliNgam']
lingas = set(_lingas)
napumsakam = _lingas[1]

avyaya = set(['avyayam'])
kriyavisheshana = set(['kriyAviSezaRam'])
nishedha = set(['na'])
karmap_2 = set(['anu', 'upa',  'prati', 'aBi', 'aDi', 'ati'])
karmap_5 = set(['apa', 'pari', 'A', 'prati'])
# FIXME - api and su (pUjAyaM) in here temporarily - need a better solution
karmap_null = set(['su', 'api'])
avyaya_kriyav = set(['kila', 'bata', 'aho', 'nanu', 'hanta', 'eva', 'tu'])
projlabels = karakas.union(kriyavisheshana)
# sambaddha links are projective
samplabels = {'sambadDa-'+l for l in projlabels}
projlabels.update(samplabels)
sentence_conjunctions = {"yad": "tad", "yadi": "tarhi", "yatra": "tatra", "yaTA": "taTA", "api": None, "cet": None, "yat": None, "natu": None}


class VakyaGraph(object):
    """ DAG class for Sanskrit Vakya Analysis

        Represents Sanskrit Vakya Analysis as a DAG
        Nodes are SanskritObjects with morphological tags
        Edges are potential relationships between them
    """
    def __init__(self, path, max_parse_dc=4, fast_merge=True):
        ''' DAG Class Init

        Params:
             path: Path from SandhiGraph
        '''
        self.max_parse_dc = max_parse_dc
        self.fast_merge = fast_merge
        self.roots = []
        self.isLocked = False
#        self.G = nx.MultDiGraph() # Allow parallel edges
        self.G = nx.DiGraph()
        # Need this many nodes in the extracted subgraphs
        self.path_node_count = len(path)
        logger.debug(f"{self.path_node_count} sets of orthogonal nodes")
        self.partitions = []
        for (ix, sobj) in enumerate(path):
            vnlist = []
            mtags = sobj.getMorphologicalTags()
            for mtag in mtags:
                ncopy = SanskritObject(sobj.canonical(),
                                       encoding=SLP1,
                                       replace_ending_visarga=None)
                ncopy.setMorphologicalTags(mtag)
                pn = VakyaGraphNode(ncopy, ix)
                vnlist.append(pn)
            for vn in vnlist:
                self.add_node(vn)
            self.partitions.append(set(vnlist))
        logger.debug(f"Node Partitions {self.partitions} Len {len(self.partitions)}")
        self.lock()
        self.add_edges()
        # Remove isolated nodes (with no edges)
        isolates = list(nx.isolates(self.G))
        self.G.remove_nodes_from(isolates)
        for (ix, s) in enumerate(self.partitions):
            s.difference_update(isolates)
            if len(s) == 0:
                logger.error(f"Partition {ix}: {path[ix]} went to zero length!")
        start_parse = time.time()
        self.parses = self.get_parses_dc()
        end_parse = time.time()
        self.check_parse_validity()
        end_check = time.time()
        logger.info(f"Time for parse: {(end_parse-start_parse):1.6f}s")
        logger.info(f"Total for Global Check: {(end_check-end_parse):1.6f}s")
        logger.info(f"Total Time for parse: {(end_check-start_parse):1.6f}s")

    def __iter__(self):
        ''' Iterate over nodes '''
        return self.G.__iter__()

    def __str__(self):
        """ Print representation of DAG """
        return str(self.G)

    def lock(self):
        self.isLocked = True

    def add_node(self, node):
        """ Extend dag with node inserted at root

            Params:
                Node (VakyaGraphNode)      : Node to add
        """
        assert node not in self.G
        assert not self.isLocked
        self.G.add_node(node)

    def add_edges(self):
        assert self.isLocked
        laks = self.find_lakaras()
        krts = self.find_krtverbs()
        bases = []
        bases.extend(laks)
        bases.extend(krts)
        logger.debug("Adding Edges")
        self.add_karakas(bases)
        self.add_samastas()
        self.add_shashthi()
        self.add_kriyavisheshana(bases)  # FIXME Parallel edge problem
        self.add_visheshana()
        self.add_kriya_kriya(laks, krts)
        self.add_avyayas(bases)
        self.add_bhavalakshana(krts, laks)
        self.add_vipsa()
        self.add_sentence_conjunctions(bases)

    def find_krtverbs(self):
        ''' Find non ti~Nanta verbs'''
        rlist = []
        for n in self.G:
            if n.node_is_a(krtverbs):
                logger.debug(f"{n} is a possible Krt Dhatu")
                rlist.append(n)
        return rlist

    def find_lakaras(self):
        ''' Find the ti~Nanta '''
        rlist = []
        for n in self.G:
            if n.node_is_a(lakaras):
                logger.debug(f"{n} is a possible Dhatu")
                rlist.append(n)
        return rlist

    def add_visheshana(self):
        for n in self.G:
            if n.node_is_a(vibhaktis):
                for no in self.G:
                    if (not _is_same_partition(n, no)) and match_linga_vacana_vibhakti(n, no):
                        if _get_base(n) != _get_base(no):
                            logger.debug(f"Adding viSezaRa edge: {n,no}")
                            self.G.add_edge(n, no, label="viSezaRa")

    def add_vipsa(self):
        for n in self.G:
            for no in self.G:
                if (n.index == (no.index-1)) and \
                   (n.pada.canonical() == no.pada.canonical()):
                    logger.debug(f"Adding vIpsa edge: {n, no}")
                    self.G.add_edge(n, no, label="vIpsA")

    def add_samastas(self):
        ''' Add samasta links from next samasta/tiN '''
        for (i, s) in enumerate(self.partitions):
            for n in s:
                # If node is a samasta, check options for
                # next node, and add samasta links if tiN
                if n.node_is_a(samastas):
                    # Cant have samasta as last node
                    if i < (len(self.partitions)-1):
                        nextset = self.partitions[i+1]
                        for nn in nextset:
                            if nn.node_is_a(vibhaktis) or \
                               nn.node_is_a(samastas):
                                logger.debug(f"Adding samasta edge: {n,nn}")
                                self.G.add_edge(nn, n, label="samasta")

    def add_shashthi(self):
        ''' Add zazWI-sambanDa links to next tiN '''
        for (i, s) in enumerate(self.partitions):
            for n in s:
                # If node is a shashthi, check
                # next node, and add links if tiN
                if n.node_is_a(shashthi):
                    # Cant have sambandha open at last node
                    if i < (len(self.partitions)-1):
                        nextset = self.partitions[i+1]
                        for nn in nextset:
                            if nn.node_is_a(vibhaktis) or \
                               nn.node_is_a(samastas):
                                logger.debug(f"Adding shashthi-sambandha edge: {n,nn}")
                                self.G.add_edge(nn, n, label="zazWI-sambanDa")

    def add_karakas(self, bases):
        ''' Add karaka edges from base node (dhatu) base '''
        for d in bases:
            logger.debug(f"Processing {d}")
            dh = _get_base(d).canonical()
            hpos = dh.find("#")
            if hpos != -1:
                dh = dh[:hpos]
            if d.node_is_a(lakaras) or d.node_is_a('avyayaDAturUpa'):
                is_sak = dw.is_sakarmaka(dh)
            else:
                is_sak = True  # No way of knowing, set True
            logger.debug(f"Dhatu: {dh} Sakarmaka {is_sak}")
            if d.node_is_a(karmani):
                logger.debug("Karmani")
                karta = tritiya
                karma = prathama
            else:
                logger.debug("Kartari")
                karta = prathama
                karma = dvitiya
            for n in self.G:
                if not _is_same_partition(d, n):
                    if n.node_is_a(karta):
                        # Only Lakaras and karmani krts are allowed kartA
                        if d.node_is_a(lakaras):
                            if match_purusha_vacana(d, n):
                                logger.debug(f"Adding kartA edge to {n}")
                                self.G.add_edge(d, n, label="kartA")
                            elif dh in predicative_verbs:
                                logger.debug(f"Adding kartfsamAnADikaraRa edge to {n}")
                                self.G.add_edge(d, n, label="kartfsamAnADikaraRa")
                        elif d.node_is_a(karmani) and match_linga_vacana(d, n):
                            logger.debug(f"Adding kartA edge to {n}")
                            self.G.add_edge(d, n, label="kartA")

                    elif (n.node_is_a(karma) and
                          (d.node_is_a(lakaras) or not d.node_is_a(karmani))
                          and is_sak):
                        # Likewise, only Lakaras and kartari krts are allowed
                        # Karma
                        logger.debug(f"Adding karma edge to {n}")
                        self.G.add_edge(d, n, label="karma")
                    elif n.node_is_a(tritiya):
                        logger.debug(f"Adding karana edge to {n}")
                        self.G.add_edge(d, n, label="karaRam")
                    elif n.node_is_a(chaturthi):
                        logger.debug(f"Adding sampradana edge to {n}")
                        self.G.add_edge(d, n, label="sampradAnam")
                    elif n.node_is_a(pancami):
                        logger.debug(f"Adding apadana edge to {n}")
                        self.G.add_edge(d, n, label="apAdanam")
                    elif n.node_is_a(saptami):
                        logger.debug(f"Adding adhikarana edge to {n}")
                        self.G.add_edge(d, n, label="aDikaraRam")
                    elif n.node_is_a(sambodhana) and check_sambodhya(d, n):
                        logger.debug(f"Adding sambodhya edge to {n}")
                        self.G.add_edge(d, n, label="samboDyam")

    def add_kriyavisheshana(self, bases):
        ''' Add kriyaviSezaRa edges from base node (dhatu) base '''
        for d in bases:
            for n in self.G:
                if not _is_same_partition(d, n):
                    if n.node_is_a(avyaya) and \
                         (n.node_is_a(kriyavisheshana) or
                          _get_base(n) in avyaya_kriyav):
                        logger.debug(f"Adding kriyAviSezaRa edge to {n}")
                        self.G.add_edge(d, n, label="kriyAviSezaRam")

    def add_kriya_kriya(self, lakaras, krts):
        ''' Add kriya-kriya edges from lakaras to krts'''
        for d in lakaras:
            for n in krts:
                if not _is_same_partition(d, n):
                    if n.node_is_a(purvakala):
                        logger.debug(f"Adding pUrvakAlaH edge to {n}")
                        self.G.add_edge(d, n, label="pUrvakAlaH")
                    elif n.node_is_a('tumun'):
                        logger.debug(f"Adding prayojanam edge to {n}")
                        self.G.add_edge(d, n, label="prayojanam")
                    elif n.node_is_a(samanakala) and n.node_is_a(prathama):
                        logger.debug(f"Adding samAnakAlaH edge to {n}")
                        self.G.add_edge(d, n, label="samAnakAlaH")

    def add_avyayas(self, bases):
        ''' Add Avyaya Links '''
        # Upasargas
        for (i, s) in enumerate(self.partitions):
            for n in s:
                if n.node_is_a('upasargaH'):
                    # Cant have upasargas at last node
                    if i < (len(self.partitions)-1):
                        nextset = self.partitions[i+1]
                        # Upasarga to upasarga links ok
                        # Upasarga to any verb form barring ktvA
                        for nn in nextset:
                            if ((nn in bases) and not nn.node_is_a('ktvA')) or \
                               nn.node_is_a('upasargaH'):
                                logger.debug(f"Adding upasarga edge: {n,nn}")
                                self.G.add_edge(nn, n, label="upasargaH")
                elif n.node_is_a(avyaya) and (_get_base(n) in nishedha):
                    for b in bases:
                        if not _is_same_partition(n, b):
                            logger.debug(f"Adding nishedha edge: {n, b}")
                            self.G.add_edge(b, n, label="nizeDa")
                elif n.node_is_a('karmapravacanIyaH') and not (_get_base(n) in avyaya_kriyav) and not(_get_base(n) in karmap_null):
                    for b in bases:
                        if not _is_same_partition(n, b):
                            logger.debug(f"Adding karmapravacaniya karma edge: {n, b}")
                            self.G.add_edge(b, n, label="karma")
                    nextset = self.partitions[i+1]
                    prevset = self.partitions[i-1]
                    for nn in nextset.union(prevset):
                        if nn.node_is_a(dvitiya) and (_get_base(n) in karmap_2):
                            logger.debug(f"Adding karmapravacaniya upapada 2 edge: {n,nn}")
                            self.G.add_edge(n, nn, label="upapadadvitIya")
                        elif nn.node_is_a(pancami) and (_get_base(n) in karmap_5):
                            logger.debug(f"Adding karmapravacaniya upapada 5 edge: {n,nn}")
                            self.G.add_edge(n, nn, label="upapadapancami")

    def add_bhavalakshana(self, krts, laks):
        ''' Add bhavalakshana edges from saptami krts to lakaras '''
        for k in krts:
            if k.node_is_a(saptami):
                for l in laks:
                    if not _is_same_partition(k, l):
                        logger.debug(f"Adding Bhavalakshana edge: {k, l}")
                        self.G.add_edge(l, k, label="BAvalakzaRam")

    def add_sentence_conjunctions(self, bases):
        ''' Add sentence conjunction links

        For all nodes which match sentence_conjuction keys
        add vAkyasambanDaH link between y-t pair
              - if vibhakti matches where relevant
        Reverse all edges to node, add sambadDa- to link label (eg sambadDa-karma, if node is not vIpsa
        if node is saMyojakaH, and not vIpsA add saMbadDakriyA links to verbs
        if associated t* doesn't exist vAkyasambanDaH links from verbs
        '''
        def _is_vipsa(n):
            for p, d in self.G.pred[n].items():
                if d['label'] == 'vIpsA':
                    return True
            return False

        sentence_conjunctions_y = set(sentence_conjunctions.keys())
        for n in self.G:
            nb = _get_base(n)
            if nb in sentence_conjunctions_y:
                if not _is_vipsa(n):
                    # Reverse edges, add sambadDa label
                    for p, d in list(self.G.pred[n].items()):
                        self.G.remove_edge(p, n)
                        self.G.add_edge(n, p, label='sambadDa-'+d['label'])
                    # Matching pair
                    s_t = sentence_conjunctions[nb]
                    for nn in self.G:
                        if (not _is_vipsa(nn)) and _get_base(nn) == s_t:
                            if match_linga_vacana(n, nn):
                                logger.info(f"Adding vAkyasambanDaH edge: {nn, n}")
                                self.G.add_edge(nn, n, label="vAkyasambanDaH")
                        if n.node_is_a('saMyojakaH') and (nn in bases):
                            logger.info(f"Adding saMbadDakriyA edge: {nn, n}")
                            self.G.add_edge(n, nn, label='saMbadDakriyA')
                            if s_t is None:
                                logger.info(f"Adding vAkyasambanDaH edge: {nn, n}")
                                self.G.add_edge(nn, n, label="vAkyasambanDaH")

    def get_parses_dc(self):
        ''' Returns all parses

            Uses modified Kruskal Algorithm to compute (generalized) spanning
            tree of k-partite VakyaGraph
        '''
        logger.debug("Computing Parses (Divide & Conquer)")

        def _get_parse_sub(mn, mx):
            # Iterate over subsets of disjoint nodesets
            for (i, ns) in enumerate(islice(self.partitions, mn, mx)):
                logger.debug(f"Node set number {i} {ns}")
                if i == 0:
                    # Partial parses = phi + all predecessors
                    partial_parses = set()
                    partial_parses.add(VakyaParse(None))  # Null partial parse
                    # For all input edges to this set
                    for n in ns:
                        logger.debug(f"Traversing node {n}")
                        for pred in self.G.predecessors(n):
                            logger.debug(f"Traversing predecessor {pred} -> {n}")
                            partial_parses.add(VakyaParse((pred, n)))
                else:
                    store_parses = set()
                    small_parses = set()
                    for ps in partial_parses:  # For each partial parse
                        if len(ps) < i:  # Small parses to be removed
                            small_parses.add(ps)
                        for n in ns:  # For all input edges to this set
                            logger.debug(f"Traversing node {n}")
                            for pred in self.G.predecessors(n):
                                logger.debug(f"Traversing predecessor {pred} -> {n}")
                                # If edge is compatible with partial parse, add and create new partial parse
                                logger.debug(f"Trying to extend parse {ps}")
                                if ps.is_safe(pred, n):
                                    logger.debug(f"{pred} - {n} is safe for {ps}")
                                    psc = ps.copy()  # Copy the nodeset and DisjointSet structures
                                    psc.extend(pred, n)
                                    store_parses.add(psc)
                    partial_parses.difference_update(small_parses)
                    partial_parses.update(store_parses)
                logger.debug(f"Partial Parses")
                for p in partial_parses:
                    logger.debug(p)
            return partial_parses

        # Divide & Conquer routine
        def _dc(mn, mx):
            logger.info(f"Divide And Conquer {mn, mx}")
            if (mx - mn) > self.max_parse_dc:
                md = int((mx + mn)/2)
                return _merge_partials(_dc(mn, md), _dc(md, mx), mx, mn)
            else:
                t = _get_parse_sub(mn, mx)
                return t

        def _merge_partials(pp1, pp2, mx, mn):
            logger.info(f"Merging between {mn, mx} {len(pp1)} x {len(pp2)}")
            start_time = time.time()
            logger.debug(f"Merging {pp1, pp2}")
            ppmt = set()
            for ppa in pp1:
                for ppb in pp2:
                    if self.fast_merge:
                        if ppa.can_merge(ppb, mx-mn-1):
                            ppmt.add(ppa.merge_f(ppb))
                    else:  # Slow Merge
                        merged = ppa.merge_s(ppb, mx-mn-1)
                        if merged:
                            ppmt.add(merged)
            logger.debug(f"{len(ppmt)} parses")
            logger.debug(f"Merged {ppmt}")
            end_time = time.time()
            logger.info(f"Time for merge {end_time-start_time}")
            return ppmt

        partial_parses = _dc(0, self.path_node_count)
        logger.debug(f"Partial Parses Before Return {partial_parses}")
        return set([self.G.edge_subgraph(p.edges) for p in partial_parses])

    def check_parse_validity(self):
        ''' Validity Check for parses

            Remove parses with double kArakas
            Remove parses with multiple to edges into a node
            Remove parses with cycles
        '''
        iv = set()
        logger.debug(f"Parses before validity check {len(self.parses)}")
        for p in self.parses:
            if not _check_parse(p):
                logger.debug(f"Will remove {p}")
                iv.add(p)  # Collect invalid parses
        # Remove them
        self.parses.difference_update(iv)
        logger.info(f"Parses after validity check {len(self.parses)}")

    def draw(self, *args, **kwargs):
        _ncache = {}

        def _uniq(s):
            if s not in _ncache:
                _ncache[s] = 0
                return s
            else:
                _ncache[s] = _ncache[s] + 1
                return s + "_" + str(_ncache[s])

        nx.draw(self.G, *args, **kwargs,
                pos=nx.graph_layout(self),
                labels={x: _uniq(str(x)) for x in self})

    def write_dot(self, path):
        nx.drawing.nx_pydot.write_dot(self.G, path)
        d = dirname(path)
        be = basename(path)
        b, e = splitext(be)
        logger.debug(f"Path {d} {b} {e}")
        for i, p in enumerate(self.parses):
            pt = join(d, b+f"_parse{i}"+e)
            nx.drawing.nx_pydot.write_dot(p, pt)


class VakyaGraphNode(object):
    """ Class for VakyaGraph nodes

        This has pada (a SanskritObject) plus a list of disjoint nodes
        to which edges cannot be created from this node
    """
    def __init__(self, sobj, index):
        self.pada = sobj
        self.index = index

    def getMorphologicalTags(self):
        return self.pada.getMorphologicalTags()

    def getNodeTagset(self):
        ''' Given a Node, extract the tagset '''
        return getSLP1Tagset(self.getMorphologicalTags())

    def node_is_a(self, st):
        ''' Check if node matches a particular tag or any of a set of tags '''
        if isinstance(st, str):
            return st in self.getNodeTagset()
        elif isinstance(st, set):
            return not st.isdisjoint(self.getNodeTagset())
        else:
            logger.error(f"node_is_a: expecting str or set, got {st}")

    def get_vibhakti(self):
        ''' Get Node vacana '''
        return self.getNodeTagset().intersection(vibhaktis)

    def get_vacana(self):
        ''' Get Node vacana '''
        return self.getNodeTagset().intersection(vacanas)

    def get_linga(self):
        ''' Get Node linga '''
        return self.getNodeTagset().intersection(lingas)

    def get_purusha(self):
        ''' Get Node puruza '''
        return self.getNodeTagset().intersection(puruzas)

    def get_node_pos(self, node):
        return self.path.index(node.pada.canonical())

    def __str__(self):
        return str(self.pada) + "=>" + str(self.pada.getMorphologicalTags()) + \
            ": " + str(self.index)

    def __repr__(self):
        return str(self)


class VakyaParse(object):
    def __init__(self, nodepair):
        ''' Initializes a partial parse with a node pair (or []) '''
        # DisjointSet  (as in Kruskal)
        self.connections = DisjointSet()
        # "Extinguished" nodes - nodes from partitions whose representatives
        # Have been added to the forest/ST already
        self.extinguished = set()
        # Nodes in the forest/ST
        self.activenodes = set()
        if nodepair is not None:
            self._populate(nodepair)
        else:
            # Edges in the forest/ST
            self.edges = set()

    def __repr__(self):
        return str(self)

    def __str__(self):
        return str([(u.pada.canonical(), v.pada.canonical()) for (u, v) in self.edges])

    def _populate(self, nodepair):
        # Add first edge to forest
        self.edges = set([(nodepair[0], nodepair[1])])
        for n in nodepair:
            self.activate_and_extinguish_alternatives(n)
        self.connections.union(nodepair[0], nodepair[1])

    def activate_and_extinguish_alternatives(self, node):
        ''' Make node active, extinguish other nodes in its partition '''
        self.activenodes.add(node)
        self.extinguished.add(node.index)

    def is_extinguished(self, node):
        ''' Is a node extinguished '''
        return (node.index in self.extinguished) and \
            (node not in self.activenodes)

    def is_safe(self, pred, node):
        ''' Checks if a partial parse is compatible with a given node and predecessor pair '''
        if self.is_extinguished(pred) or self.is_extinguished(node):
            r = False
        elif (pred in self.activenodes) and (node in self.activenodes):
            r = not self.connections.connected(pred, node)
        else:
            r = True
        return r

    def extend(self, pred, node):
        ''' Extend current parse with edge from pred to node '''
        if not self.activenodes:
            logger.debug(f"Populating {self.edges} with {(pred,node)}")
            self._populate((pred, node))
        else:
            logger.debug(f"Extending {self.edges} with {(pred,node)}")
            if pred not in self.activenodes:
                self.activate_and_extinguish_alternatives(pred)
            if node not in self.activenodes:
                self.activate_and_extinguish_alternatives(node)
            self.edges.add((pred, node))
            self.connections.union(pred, node)
        logger.debug(f"Edges {self.edges}")

    def can_merge(self, other, l):
        ''' Can we merge two VakyaParses '''
        # Proper length
        if (len(other.edges) + len(self.edges)) < l:
            return False
        # No extinguished nodes
        for x in other.activenodes:
            if self.is_extinguished(x):
                return False
        conn = self.connections.copy()
        # No cycles
        for (u, v) in other.edges:
            if conn.connected(u, v):
                return False
            else:
                conn.union(u, v)
        return True

    def merge_f(self, other):
        ''' Merge two VakyaParses: Fast method '''
        t = self.copy()
        logger.debug("Merging")
        t.extinguished.update(other.extinguished)
        t.activenodes.update(other.activenodes)
        t.edges.update(other.edges)
        for (u, v) in other.edges:
            t.connections.union(u, v)
        return t

    def merge_s(self, other, l):
        ''' Merge two VakyaParses: Slow method '''
        # Proper length
        if (len(other.edges) + len(self.edges)) < l:
            return False
        t = self.copy()
        for (u, v) in other.edges:
            if t.is_safe(u, v):
                t.extend(u, v)
            else:
                return False
        return t

    def __len__(self):
        return len(self.edges)

    def copy(self):
        ''' Return a one level deep copy - in between a shallow and a fully deep copy '''
        t = VakyaParse(None)
        t.activenodes = copy(self.activenodes)
        t.edges = copy(self.edges)
        t.extinguished = copy(self.extinguished)
        t.connections = self.connections.copy()
        return t


def getSLP1Tagset(n):
    ''' Given a (base, tagset) pair, extract the tagset '''
    return set(map(lambda x: x.canonical(), list(n[1])))


def match_purusha_vacana(d, n):
    ''' Check vacana/puruza compatibility for a Dhatu d and node n '''
    n_base = _get_base(n)
    if n_base == 'asmad':
        n_purusha = set([puruzas[2]])
    elif n_base == 'yuzmad':
        n_purusha = set([puruzas[1]])
    else:
        n_purusha = set([puruzas[0]])
    return (d.get_vacana() == n.get_vacana()) and (d.get_purusha() == n_purusha)


def match_linga_vacana(n1, n2):
    ''' Check linga/puruza compatibility for two nodes '''
    return (n1.get_vacana() == n2.get_vacana()) and \
        (n1.get_linga() == n2.get_linga())


def match_linga_vacana_vibhakti(n1, n2):
    return (n1.get_vacana() == n2.get_vacana()) and \
        (n1.get_linga() == n2.get_linga()) and \
        (n1.get_vibhakti() == n2.get_vibhakti())


def check_sambodhya(d, n):
    ''' Check sambodhya compatibility for dhatu d and node n '''
    return (d.get_vacana() == n.get_vacana()) and \
        (d.get_purusha() == set([puruzas[1]]))


def jedge(pred, node, label, strict_io=False):
    return (node.pada.canonical(strict_io=strict_io),
            jtag(node.getMorphologicalTags(), strict_io),
            SanskritImmutableString(label, encoding=SLP1).canonical(strict_io=strict_io),
            pred.pada.canonical(strict_io=strict_io))


def jnode(node, strict_io=False):
    """ Helper to translate parse node into serializable format"""
    return (node.pada.canonical(strict_io=strict_io),
            jtag(node.getMorphologicalTags(), strict_io), "", "")


def jtag(tag, strict_io=False):
    """ Helper to translate tag to serializable format"""
    return (tag[0].canonical(strict_io=strict_io), [t.canonical(strict_io=strict_io) for t in list(tag[1])])


def _non_projective(u, v, w, x):
    """ Checks if an edge pair is non-projective """
    mnu = min(u, v)
    mxu = max(u, v)
    mnw = min(w, x)
    mxw = max(w, x)
    if mnu < mnw:
        return (mxu < mxw) and (mxu > mnw)
    elif mxu > mxw:
        return (mnu > mnw) and (mnu < mxw)


def _is_same_partition(n1, n2):
    ''' Are the two nodes n1 and n2 are in the same partition in psets '''
    return n1.index == n2.index


def _get_base(n):
    return n.getMorphologicalTags()[0]


# Check a parse for validity
def _check_parse(parse):
    r = True
    smbds = samplabels.union({'saMbadDakriyA'})
    count = defaultdict(lambda: defaultdict(int))
    edges = {}
    toedge = defaultdict(int)
    fromv = defaultdict(int)
    tov = defaultdict(int)
    sk = defaultdict(int)
    for (u, v, l) in parse.edges(data='label'):
        if l in karakas:
            count[u][l] = count[u][l]+1
            toedge[v] = toedge[v]+1
        if l in 'viSezaRa':
            fromv[u] = fromv[u] + 1
            tov[v] = tov[v] + 1
        if l in projlabels:  # Labels with sannidhi expectation
            edges[(u.index, v.index)] = 1
        if l in smbds:
            sk[u] = sk[u] + 1
    for u in count:  # Dhatu
        for k in count[u]:  # Each karaka should count only once
            if count[u][k] > 1:
                logger.debug(f"Count for {u} {k} is {count[u][k]} - violates global constraint")
                return False
    for (ui, vi) in edges:
        for (wi, xi) in edges:
            if _non_projective(ui, vi, wi, xi):  # Non-projective
                logger.debug(f"Sannidhi violation {ui} - {vi} : {wi} - {xi}")
                return False
    for v in toedge:
        if toedge[v] > 1:
            logger.debug(f"Toedges for {v} is {toedge[v]} - violates global constraint")
            return False
    for u in sk:   # Sambaddhakaraka = only one allowed from node
        if sk[u] > 1:
            logger.debug(f"Sambaddha karaka edges for {u} is {sk[u]} - violates global constraint")
            return False
    for v in tov:
        if v in fromv:
            logger.debug(f"Viseshana has visheshana {v} - violates global constraint")
            return False
    return r



# -*- coding: utf-8 -*-
"""
@author: Karthik Madathil (github: @kmadathil)

"""
from os.path import dirname, basename, splitext, join
from argparse import ArgumentParser
import logging
from sanskrit_parser import Parser
from sanskrit_parser.base.sanskrit_base import SCHEMES, SanskritNormalizedString
from sanskrit_parser.base.sanskrit_base import outputctx
from sanskrit_parser.parser.sandhi_analyzer import LexicalSandhiAnalyzer
from sanskrit_parser import enable_file_logger, enable_console_logger

logger = logging.getLogger(__name__)


def getVakyaArgs(argv=None):
    """
      Argparse routine.
      Returns args variable
    """
    # Parser Setup
    parser = ArgumentParser(description='Vakya Analyzer')
    # String to encode
    parser.add_argument('data', nargs="?", type=str, default="astyuttarasyAMdishidevatAtmA")
    # Input Encoding (autodetect by default)
    parser.add_argument('--input-encoding', type=str, default=None)
    # Need a lakara
    parser.add_argument('--need-lakara', action='store_true')
    parser.add_argument('--max-paths', type=int, default=1)
    parser.add_argument('--split-above', type=int, default=5)
    parser.add_argument('--lexical-lookup', type=str, default="combined")
    parser.add_argument('--strict-io', action='store_true',
                        help="Do not modify the input/output string to match conventions", default=False)
    parser.add_argument('--score', dest="score", action='store_true',
                        help="Use the lexical scorer to score the splits and reorder them")
    parser.add_argument('--slow-merge', dest='fast_merge', action='store_false', help="Development Only: use if you see issues in divide and conquer")
    parser.add_argument('--dot-file', type=str, default=None, help='Dotfile')
    return parser.parse_args(argv)


def vakya(argv=None):
    args = getVakyaArgs(argv)
    if args.strict_io:
        print("Interpreting input strictly")
    else:
        print("Interpreting input loosely (strict_io set to false)")
    logger.info(f"Input String: {args.data}")
    parser = Parser(input_encoding=args.input_encoding,
                    strict_io=args.strict_io,
                    output_encoding="SLP1",
                    replace_ending_visarga=None,
                    score=args.score,
                    split_above=args.split_above,
                    lexical_lookup=args.lexical_lookup)
    parse_result = parser.parse(args.data)
    print('Splits:')
    logger.debug('Splits:')
    for si, split in enumerate(parse_result.splits(max_splits=args.max_paths)):
        logger.info(f'Lexical Split: {split}')
        for pi, parse in enumerate(split.parses()):
            logger.debug(f'Parse {pi}')
            logger.debug(f'{parse}')
            print(f'Parse {pi}')
            print(f'{parse}')
        # Write dot files
        if args.dot_file is not None:
            path = args.dot_file
            d = dirname(path)
            be = basename(path)
            b, e = splitext(be)
            splitbase = join(d, b + f"_split{si}" + e)
            split.write_dot(splitbase)

    return None


def getSandhiArgs(argv=None):
    """
      Argparse routine.
      Returns args variable
    """
    # Parser Setup
    parser = ArgumentParser(description='Sandhi Analyzer')
    # String to encode
    parser.add_argument('data', nargs="?", type=str, default="astyuttarasyAMdishidevatAtmA")
    # Input Encoding (autodetect by default)
    parser.add_argument('--input-encoding', type=str, default=None)
    parser.add_argument('--max-paths', type=int, default=10)
    parser.add_argument('--lexical-lookup', type=str, default="combined")
    parser.add_argument('--strict-io', action='store_true',
                        help="Do not modify the input/output string to match conventions", default=False)
    parser.add_argument('--no-score', dest="score", action='store_false',
                        help="Use the lexical scorer to score the splits and reorder them")
    parser.add_argument('--dot-file', type=str, default=None, help='Dotfile')
    return parser.parse_args(argv)


def sandhi(argv=None):
    args = getSandhiArgs(argv)
    if args.strict_io:
        print("Interpreting input strictly")
    else:
        print("Interpreting input loosely (strict_io set to false)")
    logger.info(f"Input String: {args.data}")
    parser = Parser(input_encoding=args.input_encoding,
                    strict_io=args.strict_io,
                    output_encoding="SLP1",
                    replace_ending_visarga=None,
                    score=args.score,
                    lexical_lookup=args.lexical_lookup)
    parse_result = parser.parse(args.data)
    print('Splits:')
    logger.debug('Splits:')
    for si, split in enumerate(parse_result.splits(max_splits=args.max_paths)):
        logger.info(f'Split: {split}')
    # Write dot files
    if args.dot_file is not None:
        parse_result.write_dot(args.dot_file)

    return None


def getTagsArgs(argv=None):
    """
      Argparse routine.
      Returns args variable
    """
    # Parser Setup
    parser = ArgumentParser(description='Morphological tags')
    # String to encode
    parser.add_argument('data', nargs="?", type=str, default="adhi")
    # Input Encoding (autodetect by default)
    parser.add_argument('--input-encoding', type=str, default=None)
    # Filter by base name
    parser.add_argument('--base', type=str, default=None)
    # Filter by tag set
    parser.add_argument('--tag-set', type=str, default=None, nargs="+")
    parser.add_argument('--tags', dest='split', action='store_false')
    parser.add_argument('--lexical-lookup', type=str, default="combined")
    parser.add_argument('--strict-io', action='store_true',
                        help="Do not modify the input/output string to match conventions", default=False)
    parser.add_argument('--no-map-tags', dest='map_tags',
                        action='store_false', help="show raw (unmapped to our standard set) tags")
    return parser.parse_args(argv)


def tags(argv=None):
    args = getTagsArgs(argv)
    if args.strict_io:
        print("Interpreting input strictly")
    else:
        print("Interpreting input loosely (strict_io set to false)")
    logger.info(f"Input String: {args.data}")
    if args.input_encoding is None:
        ie = None
    else:
        ie = SCHEMES[args.input_encoding]
    s = LexicalSandhiAnalyzer(args.lexical_lookup)
    with outputctx(args.strict_io):
        i = SanskritNormalizedString(args.data, encoding=ie,
                                     strict_io=args.strict_io,
                                     replace_ending_visarga='s')
        print("Input String in SLP1:", i.canonical())
        ts = s.getMorphologicalTags(i, tmap=args.map_tags)
        print("Morphological tags:")
        if ts is not None:
            for t in ts:
                print(t)
        # Possible rakaranta
        # Try by replacing end visarga with 'r' instead
        elif not args.strict_io:
            i = SanskritNormalizedString(args.data, encoding=ie,
                                         strict_io=args.strict_io,
                                         replace_ending_visarga='r')
            ts = s.getMorphologicalTags(i)
            if ts is not None:
                print("Input String in SLP1:", i.canonical())
                for t in ts:
                    print(t)
        if args.tag_set or args.base:
            if args.tag_set is not None:
                g = set(args.tag_set)
            else:
                g = None
            if args.base is not None:
                b = SanskritNormalizedString(args.base)
            else:
                b = None
            print(s.hasTag(i, b, g))


def cmd_line():
    """ Command Line Wrapper Function
    """
    parser = ArgumentParser(description='Sanskrit Parser')

    parser.add_argument('command', help='Subcommand to run',
                        choices=["sandhi", "vakya", "tags"])
    parser.add_argument('--debug', action='store_true')

    # parse_args defaults to [1:] for args, but you need to
    # exclude the rest of the args too, or validation will fail
    args, rest = parser.parse_known_args()

    # Logging
    enable_console_logger(level=logging.INFO)
    if args.debug:
        enable_file_logger(level=logging.DEBUG)
    # else:
    #    enable_file_logger(level=logging.INFO)

    if not hasattr(args, 'command'):
        print('Unrecognized command')
        parser.print_help()
        exit(1)
    # use dispatch pattern to invoke method with same name
    eval(getattr(args, 'command')+"(rest)")



'''
Base class + factory for lexical lookup classes

@author: avinashvarna
'''

from __future__ import print_function
import abc
import os
from argparse import ArgumentParser
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SCHEMES


class LexicalLookup(object):

    __metaclass__ = abc.ABCMeta

    base_dir = os.path.expanduser("~/.sanskrit_parser/data")

    @abc.abstractmethod
    def valid(self, word):
        """ Return True if word is a valid pada """

    @abc.abstractmethod
    def get_tags(self, word, tmap=True):
        """ Return lexical tags of word """

    @staticmethod
    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = ArgumentParser(description='Interface to INRIA XML database')
        # Input Encoding (autodetect by default)
        parser.add_argument('--input-encoding', type=str, default=None)
        parser.add_argument('--loglevel', type=str, default="info",
                            help="logging level. Can be any level supported by logging module")
        parser.add_argument('--no-map-tags', dest='map_tags',
                            action='store_false')

        parser.add_argument('word', nargs='?', type=str,
                            default=None,
                            help="Word to look up")

        return parser.parse_args()

    def main(self, args):
        if args.input_encoding is None:
            ie = None
        else:
            ie = SCHEMES[args.input_encoding]

        word_in = SanskritImmutableString(args.word,
                                          encoding=ie).canonical()
        print("Getting tags for", word_in)
        tags = self.get_tags(word_in, tmap=args.map_tags)
        if tags is not None:
            for tag in tags:
                print(tag)
